"""
Most datetime functions from pytesmo are now in the cadati package
Website: https://github.com/TUW-GEO/cadati
To install cadati, run 'pip install cadati'
This module will be removed in future versions of pytesmo
"""

import pandas as pd
import datetime as dt
import pytz

import warnings

from cadati.check_date import is_leap_year
from cadati.jd_date import (
    julday,
    caldat,
    julian2date,
    julian2datetime,
    julian2num,
    num2julian,
)
from cadati.conv_doy import doy
from pytesmo.utils import deprecated

# deprecated calls:
__all__ = ['is_leap_year', 'julday', 'caldat', 'julian2datetime',
           'julian2num', 'num2julian', 'doy']

warnings.warn(
    "pytesmo.timdate.julian has moved to the cadati pip package "
    "Import from pytesmo will not work in future versions of the package",
    DeprecationWarning,
)


@deprecated()
def julian2doy(j, consider_nonleap_years=True):
    """
    Calendar date from julian date.
    Works only for years past 1582!
    Parameters
    ----------
    j : numpy.ndarray or double
        Julian days.
    consider_nonleap_years : boolean, optional
        Flag if all dates are interpreted as leap years (False) or not (True).
    Returns
    -------
    doy : numpy.ndarray or int32
        Day of year.
    """
    return julian2date(
        j, return_doy=True, doy_leap_year=not consider_nonleap_years
    )[-1]


@deprecated()
def julian2datetimeindex(j, tz=pytz.UTC):
    """
    Converting Julian days to datetimeindex.
    Parameters
    ----------
    j : numpy.ndarray or int32
        Julian days.
    tz : instance of pytz, optional
        Time zone. Default: UTC
    Returns
    -------
    datetime : pandas.DatetimeIndex
        Datetime index.
    """
    year, month, day, hour, minute, second, microsecond = julian2date(j)

    return pd.DatetimeIndex(
        [
            dt.datetime(y, m, d, h, mi, s, ms, tz)
            for y, m, d, h, mi, s, ms in zip(
                year, month, day, hour, minute, second, microsecond
            )
        ]
    )



import pandas as pd
import numpy as np
import pytest

from pytesmo.validation_framework.validation import Validation
from pytesmo.validation_framework.metric_calculators import (
    PairwiseIntercomparisonMetrics,
)
from pytesmo.validation_framework.temporal_matchers import (
    make_combined_temporal_matcher,
)
import pytesmo.validation_framework.error_handling as eh
from pytesmo.validation_framework.data_manager import DataManager

from .utils import create_datasets


def test_error_handling_empty_df():
    # This tests whether error handling works if one of the datasets consists
    # of an empty dataframe.
    # The options "ignore" and "deprecated" should both return an empty
    # dictionary in this case, while the others handle the error
    # this tests whether with the "deprecated" and "ignore" option of error
    # handling, an empty dictionary is returned if one of the datasets is just
    # an empty dataframe, and therefore temporal matching fails
    n_datasets = 5
    npoints = 5
    nsamples = 100

    datasets = create_datasets(n_datasets, npoints, nsamples, missing="empty")
    metric_calculator = PairwiseIntercomparisonMetrics()

    val = Validation(
        datasets,
        spatial_ref="0-ERA5",
        metrics_calculators={(n_datasets, 2): metric_calculator.calc_metrics},
        temporal_matcher=make_combined_temporal_matcher(pd.Timedelta(12, "H")),
    )
    gpis = list(range(npoints))
    args = (gpis, gpis, gpis)
    kwargs = dict(rename_cols=False, only_with_reference=True)

    # 'raise' should raise an error
    with pytest.raises(eh.NoTempMatchedDataError):
        with pytest.warns(UserWarning, match="No data for dataset 4-missing"):
            results = val.calc(*args, **kwargs, handle_errors="raise")

    # 'ignore' should modify the status code, but not raise an error
    with pytest.warns(UserWarning, match="No data for dataset 4-missing"):
        results = val.calc(*args, **kwargs, handle_errors="ignore")
    for key in results:
        for metric in results[key]:
            assert len(results[key][metric]) == npoints
            if metric not in ["status", "gpi", "lat", "lon", "n_obs"]:
                assert np.all(np.isnan(results[key][metric]))
        assert np.all(results[key]["status"] == eh.NO_TEMP_MATCHED_DATA)


def test_error_handling_nodata():
    # this tests if we get the NoGpiDataError if one dataset doesn't have any
    # values.  Here we use only 2 datasets, otherwise the third one will be
    # available for a comparison and we will again end up with a
    # NoTempMatchedDataError
    n_datasets = 2
    npoints = 5
    nsamples = 100

    datasets = create_datasets(n_datasets, npoints, nsamples, missing="nodata")
    metric_calculator = PairwiseIntercomparisonMetrics()

    val = Validation(
        datasets,
        spatial_ref="0-ERA5",
        metrics_calculators={(n_datasets, 2): metric_calculator.calc_metrics},
        temporal_matcher=make_combined_temporal_matcher(pd.Timedelta(12, "H")),
    )
    gpis = list(range(npoints))
    args = (gpis, gpis, gpis)
    kwargs = dict(rename_cols=False, only_with_reference=True)

    # 'raise' should raise an error
    with pytest.raises(eh.NoGpiDataError):
        with pytest.warns(UserWarning, match="No data for dataset 1-missing"):
            results = val.calc(*args, **kwargs, handle_errors="raise")

    # 'ignore' should modify the status code, but not raise an error
    with pytest.warns(UserWarning, match="No data for dataset 1-missing"):
        results = val.calc(*args, **kwargs, handle_errors="ignore")
    for key in results:
        for metric in results[key]:
            assert len(results[key][metric]) == npoints
            if metric not in ["status", "gpi", "lat", "lon", "n_obs"]:
                assert np.all(np.isnan(results[key][metric]))
        assert np.all(results[key]["status"] == eh.NO_GPI_DATA)


def test_error_handling_not_enough_data():
    # This tests if we get a proper warning if we have not enough data to
    # calculate correlations (nsamples = 5). In this case, the behaviour of all
    # options should be the same (setting the correct status, but nothing
    # otherwise)
    n_datasets = 5
    npoints = 5
    nsamples = 5

    datasets = create_datasets(n_datasets, npoints, nsamples)
    metric_calculator = PairwiseIntercomparisonMetrics()

    val = Validation(
        datasets,
        spatial_ref="0-ERA5",
        metrics_calculators={(n_datasets, 2): metric_calculator.calc_metrics},
        temporal_matcher=make_combined_temporal_matcher(pd.Timedelta(12, "H")),
    )
    gpis = list(range(npoints))
    args = (gpis, gpis, gpis)
    kwargs = dict(rename_cols=False, only_with_reference=True)

    for handle_errors in ["ignore", "raise"]:
        with pytest.warns(
            UserWarning, match="Not enough observations to calculate metrics."
        ):
            results = val.calc(*args, **kwargs, handle_errors=handle_errors)
        for key in results:
            for metric in results[key]:
                assert len(results[key][metric]) == npoints
                if metric not in ["status", "gpi", "lat", "lon", "n_obs"]:
                    assert np.all(np.isnan(results[key][metric]))
            assert np.all(results[key]["status"] == eh.INSUFFICIENT_DATA)


def test_error_handling_ok():
    # everything should work out nicely here
    n_datasets = 5
    npoints = 5
    nsamples = 100

    datasets = create_datasets(n_datasets, npoints, nsamples)
    metric_calculator = PairwiseIntercomparisonMetrics()

    val = Validation(
        datasets,
        spatial_ref="0-ERA5",
        metrics_calculators={(n_datasets, 2): metric_calculator.calc_metrics},
        temporal_matcher=make_combined_temporal_matcher(pd.Timedelta(12, "H")),
    )
    gpis = list(range(npoints))
    args = (gpis, gpis, gpis)
    kwargs = dict(rename_cols=False, only_with_reference=True)

    for handle_errors in ["ignore", "raise"]:
        results = val.calc(*args, **kwargs, handle_errors=handle_errors)
        for key in results:
            for metric in results[key]:
                assert len(results[key][metric]) == npoints
                if metric not in ["status", "gpi", "lat", "lon", "n_obs"]:
                    assert np.all(np.isfinite(results[key][metric]))
            assert np.all(results[key]["status"] == eh.OK)


def test_error_handling_scaling_failed():
    # This tests whether a scaling error is raised if the scaling fails due to
    # insufficient data.
    n_datasets = 5
    npoints = 5
    nsamples = 100

    class BadScaler:
        def scale(self, data, ref_idx, gpi_info):
            raise ValueError("This is a test.")

    datasets = create_datasets(n_datasets, npoints, nsamples)
    metric_calculator = PairwiseIntercomparisonMetrics()

    val = Validation(
        datasets,
        scaling=BadScaler(),
        spatial_ref="0-ERA5",
        metrics_calculators={(n_datasets, 2): metric_calculator.calc_metrics},
        temporal_matcher=make_combined_temporal_matcher(pd.Timedelta(12, "H")),
    )
    gpis = list(range(npoints))
    args = (gpis, gpis, gpis)
    kwargs = dict(rename_cols=False, only_with_reference=True)

    # test if raising raises
    with pytest.raises(eh.ScalingError, match="Scaling failed"):
        results = val.calc(*args, **kwargs, handle_errors="raise")

    # test if ignore returns code
    results = val.calc(*args, **kwargs, handle_errors="ignore")
    for key in results:
        for metric in results[key]:
            assert len(results[key][metric]) == npoints
            if metric not in ["status", "gpi", "lat", "lon", "n_obs"]:
                assert np.all(np.isnan(results[key][metric]))
        assert np.all(results[key]["status"] == eh.SCALING_FAILED)


def test_error_handling_datamanager_failed():
    # This tests whether a scaling error is raised if the scaling fails due to
    # insufficient data.
    n_datasets = 5
    npoints = 5
    nsamples = 100

    datasets = create_datasets(n_datasets, npoints, nsamples)

    spatial_ref = "0-ERA5"
    data_manager = DataManager(datasets, spatial_ref, None)

    def bad_get_data(*args):
        raise ValueError("This is a test.")

    data_manager.get_data = bad_get_data

    metric_calculator = PairwiseIntercomparisonMetrics()

    val = Validation(
        data_manager,
        spatial_ref=spatial_ref,
        metrics_calculators={(n_datasets, 2): metric_calculator.calc_metrics},
        temporal_matcher=make_combined_temporal_matcher(pd.Timedelta(12, "H")),
    )
    gpis = list(range(npoints))
    args = (gpis, gpis, gpis)
    kwargs = dict(rename_cols=False, only_with_reference=True)

    # test if raising raises
    with pytest.raises(eh.DataManagerError, match="Getting the data for gpi"):
        results = val.calc(*args, **kwargs, handle_errors="raise")

    # test if ignore returns code
    results = val.calc(*args, **kwargs, handle_errors="ignore")
    for key in results:
        for metric in results[key]:
            assert len(results[key][metric]) == npoints
            if metric not in ["status", "gpi", "lat", "lon", "n_obs"]:
                assert np.all(np.isnan(results[key][metric]))
        assert np.all(results[key]["status"] == eh.DATA_MANAGER_FAILED)


def test_error_handling_temp_matching_failed():
    # This tests whether a TemporalMatchingError is raised if the matching
    # fails
    n_datasets = 5
    npoints = 5
    nsamples = 100

    def bad_matching(*args, **kwargs):
        raise ValueError("This is a test.")

    datasets = create_datasets(n_datasets, npoints, nsamples)
    metric_calculator = PairwiseIntercomparisonMetrics()

    val = Validation(
        datasets,
        spatial_ref="0-ERA5",
        metrics_calculators={(n_datasets, 2): metric_calculator.calc_metrics},
        temporal_matcher=bad_matching,
    )
    gpis = list(range(npoints))
    args = (gpis, gpis, gpis)
    kwargs = dict(rename_cols=False, only_with_reference=True)

    # raise should raise a TemporalMatchingError
    with pytest.raises(eh.TemporalMatchingError,
                       match="Temporal matching failed"):
        results = val.calc(*args, **kwargs, handle_errors="raise")

    # ignore should just log the correct return code
    results = val.calc(*args, **kwargs, handle_errors="ignore")
    for key in results:
        for metric in results[key]:
            assert len(results[key][metric]) == npoints
            if metric not in ["status", "gpi", "lat", "lon", "n_obs"]:
                assert np.all(np.isnan(results[key][metric]))
        assert np.all(results[key]["status"] == eh.TEMPORAL_MATCHING_FAILED)


def test_error_handling_metrics_calculation_failed():
    # This tests whether a MetricsCalculationError is raised if metrics
    # calculation fails
    n_datasets = 5
    npoints = 5
    nsamples = 100

    def bad_metrics(data, gpi_info):
        if len(data) == 0:
            return {"status": np.int32([-1])}
        raise ValueError("This is a test.")

    datasets = create_datasets(n_datasets, npoints, nsamples)

    val = Validation(
        datasets,
        spatial_ref="0-ERA5",
        metrics_calculators={(n_datasets, 2): bad_metrics},
        temporal_matcher=make_combined_temporal_matcher(pd.Timedelta(12, "H")),
    )
    gpis = list(range(npoints))
    args = (gpis, gpis, gpis)
    kwargs = dict(rename_cols=False, only_with_reference=True)

    # raise should raise the MetricsCalculationError
    with pytest.raises(eh.MetricsCalculationError,
                       match="Metrics calculation failed"):
        results = val.calc(*args, **kwargs, handle_errors="raise")

    # ignore should just log the correct return code
    results = val.calc(*args, **kwargs, handle_errors="ignore")
    for key in results:
        for metric in results[key]:
            assert len(results[key][metric]) == npoints
        assert np.all(results[key]["status"] == eh.METRICS_CALCULATION_FAILED)



# Copyright (c) 2015, Vienna University of Technology,
# Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Vienna University of Technology,
#     Department of Geodesy and Geoinformation nor the
# names of its contributors may be used to endorse or promote products #
# derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import numpy as np
import numpy.testing as nptest
import pytesmo.metrics as met
from pytesmo.metrics import *
import pytest

from pytesmo.metrics.deprecated import (
    tcol_error,
    tcol_snr,
    pearsonr,
    spearmanr,
    kendalltau,
)

from .test_metrics import arange_testdata


def assert_deprecation(testfunc):
    def new_testfunc(*args, **kwargs):
        with pytest.deprecated_call():
            return testfunc(*args, **kwargs)
    return new_testfunc


def test_deprecation_warnings():
    x = np.random.randn(100)
    y = np.random.randn(100)
    z = np.random.randn(100)

    with pytest.deprecated_call():
        rmsd(x, y, ddof=1)

    with pytest.deprecated_call():
        tcol_error(x, y, z)

    with pytest.deprecated_call():
        tcol_snr(x, y, z)

    with pytest.deprecated_call():
        pearsonr(x, y)

    with pytest.deprecated_call():
        spearmanr(x, y)

    with pytest.deprecated_call():
        kendalltau(x, y)


@assert_deprecation
def test_pearson_conf():
    """
    Test the person confidence interval based on
    the fisher z-transform
    """
    # first example
    n = 34
    r = 0.654
    rl, ru = met.pearson_conf(r, n, c=95)
    nptest.assert_almost_equal(rl, 0.406, decimal=3)
    nptest.assert_almost_equal(ru, 0.812, decimal=3)

    rl, ru = met.pearson_conf(r, n, c=99)
    nptest.assert_almost_equal(rl, 0.309, decimal=3)
    nptest.assert_almost_equal(ru, 0.8468, decimal=3)

    # second example
    r = 0.824
    n = 300
    rl, ru = met.pearson_conf(r, n, c=95)
    nptest.assert_almost_equal(rl, 0.784, decimal=3)
    nptest.assert_almost_equal(ru, 0.857, decimal=3)

    rl, ru = met.pearson_conf(r, n, c=99)
    nptest.assert_almost_equal(rl, 0.7697, decimal=3)
    nptest.assert_almost_equal(ru, 0.866, decimal=3)

    # test numpy arrays as input

    r = np.array([0.654, 0.824])
    n = np.array([34, 300])
    rl, ru = met.pearson_conf(r, n, c=95)
    nptest.assert_almost_equal(rl, np.array([0.406, 0.784]), decimal=3)
    nptest.assert_almost_equal(ru, np.array([0.812, 0.857]), decimal=3)


@assert_deprecation
def test_pearson_recursive():

    x = np.random.rand(100)
    y = np.random.rand(100)

    r, p = met.pearsonr(x, y)
    r_rec, _ = met.pearsonr_recursive(x, y)
    nptest.assert_almost_equal(r, r_rec)

    args = []
    for xi, yi in zip(x, y):

        r_rec, args = met.pearsonr_recursive(np.array([xi]),
                                             np.array([yi]), *args)

    nptest.assert_almost_equal(r, r_rec)


def test_mse(arange_testdata):
    """
    Test for mse
    """
    with pytest.deprecated_call():
        # example 1
        x, y = arange_testdata

        mse_pred = 4.
        mse_bias_pred = 2. ** 2
        mse_obs, _, mse_bias, _ = met.mse(x, y)

        nptest.assert_almost_equal(mse_obs, mse_pred)
        nptest.assert_almost_equal(mse_bias, mse_bias_pred)

        # example 2, with outlier
        y[-1] = 51.

        mse_pred = 180.
        mse_bias_pred = 36.
        mse_obs, _, mse_bias, _ = met.mse(x, y)

        nptest.assert_almost_equal(mse_obs, mse_pred, 6)
        nptest.assert_almost_equal(mse_bias, mse_bias_pred, 6)


@assert_deprecation
def test_rmsd_mse():
    """
    Test for rmsd and mse
    """
    # example 1
    x = np.random.randn(1000)
    y = np.random.randn(1000)

    rmsd_pred = met.rmsd(x, y)
    mse_pred, _, _, _ = met.mse(x, y)

    nptest.assert_almost_equal(rmsd_pred ** 2, mse_pred, 6)


@assert_deprecation
def test_tcol_error():
    """
    Test the triple collocation error estimation based on
    a random signal and error.
    Also compare the results to the other method
    """

    n = 1000000

    signal = np.sin(np.linspace(0, 2 * np.pi, n))

    sig_err_x = 0.02
    sig_err_y = 0.07
    sig_err_z = 0.04
    err_pred = np.array((sig_err_x, sig_err_y, sig_err_z))
    err_x = np.random.normal(0, sig_err_x, n)
    err_y = np.random.normal(0, sig_err_y, n)
    err_z = np.random.normal(0, sig_err_z, n)

    alpha_y = 0.2
    alpha_z = 0.5

    beta_y = 0.9
    beta_z = 1.6

    x = signal + err_x
    y = alpha_y + beta_y * (signal + err_y)
    z = alpha_z + beta_z * (signal + err_z)

    snr, err, beta = met.tcol_snr(x, y, z, ref_ind=0)
    # classical triple collocation errors use scaled (removed alpha and beta)
    # input arrays
    ex, ey, ez = met.tcol_error(signal + err_x, signal + err_y, signal + err_z)

    nptest.assert_almost_equal(err, np.array([ex, ey, ez]), decimal=2)
    nptest.assert_almost_equal(err_pred, np.array([ex, ey, ez]), decimal=2)


@assert_deprecation
def test_tcol_snr():
    """
    Test the triple collocation based estimation of
    signal to noise ratio, absolute errors and rescaling coefficients
    """

    n = 1000000

    mean_signal = 0.3
    sig_signal = 0.2
    signal = np.random.normal(mean_signal, sig_signal, n)

    sig_err_x = 0.02
    sig_err_y = 0.07
    sig_err_z = 0.04
    err_x = np.random.normal(0, sig_err_x, n)
    err_y = np.random.normal(0, sig_err_y, n)
    err_z = np.random.normal(0, sig_err_z, n)

    alpha_y = 0.2
    alpha_z = 0.5

    beta_y = 0.9
    beta_z = 1.6

    x = signal + err_x
    y = alpha_y + beta_y * (signal + err_y)
    z = alpha_z + beta_z * (signal + err_z)

    beta_pred = 1. / np.array((1, beta_y, beta_z))
    err_pred = np.array((sig_err_x, sig_err_y, sig_err_z))
    snr_pred = np.array(
        ((sig_signal / sig_err_x), (sig_signal / sig_err_y), (sig_signal / sig_err_z)))

    snr, err, beta = met.tcol_snr(x, y, z, ref_ind=0)

    nptest.assert_almost_equal(beta, beta_pred, decimal=2)
    nptest.assert_almost_equal(err, err_pred, decimal=2)
    nptest.assert_almost_equal(np.sqrt(10 ** (snr / 10.)), snr_pred, decimal=1)



"""
Old metrics implementations that have been deprecated.
"""

import numpy as np
from scipy import stats
from scipy.special import betainc

from pytesmo.utils import deprecated
from pytesmo.metrics.pairwise import bias


__all__ = []


@deprecated()
def tcol_error(x, y, z):
    """
    DEPRECATED: Use ``pytesmo.metrics.tcol_metrics`` instead.

    Triple collocation error estimate of three calibrated/scaled
    datasets.

    Parameters
    ----------
    x : numpy.ndarray
        1D numpy array to calculate the errors
    y : numpy.ndarray
        1D numpy array to calculate the errors
    z : numpy.ndarray
        1D numpy array to calculate the errors

    Returns
    -------
    e_x : float
        Triple collocation error for x.
    e_y : float
        Triple collocation error for y.
    e_z : float
        Triple collocation error for z.

    Notes
    -----
    This function estimates the triple collocation error based
    on already scaled/calibrated input data. It follows formula 4
    given in [Scipal2008]_.

    .. math:: \\sigma_{\\varepsilon_x}^2 = \\langle (x-y)(x-z) \\rangle

    .. math:: \\sigma_{\\varepsilon_y}^2 = \\langle (y-x)(y-z) \\rangle

    .. math:: \\sigma_{\\varepsilon_z}^2 = \\langle (z-x)(z-y) \\rangle

    where the :math:`\\langle\\rangle` brackets mean the temporal mean.

    References
    ----------
    .. [Scipal2008] Scipal, K., Holmes, T., De Jeu, R., Naeimi, V., Wagner,
       W. (2008). A possible solution for the problem of estimating the error
       structure of global soil moisture data sets. Geophysical Research
       Letters, 35(24), .
    """
    e_x = np.sqrt(np.abs(np.mean((x - y) * (x - z))))
    e_y = np.sqrt(np.abs(np.mean((y - x) * (y - z))))
    e_z = np.sqrt(np.abs(np.mean((z - x) * (z - y))))

    return e_x, e_y, e_z


@np.errstate(invalid="ignore")
@deprecated()
def tcol_snr(x, y, z, ref_ind=0):
    """
    DEPRECATED: Use the function `tcol_metrics` instead.

    Triple collocation based estimation of signal-to-noise ratio, absolute
    errors, and rescaling coefficients

    Parameters
    ----------
    x: 1D numpy.ndarray
        first input dataset
    y: 1D numpy.ndarray
        second input dataset
    z: 1D numpy.ndarray
        third input dataset
    ref_ind: int
        Index of reference data set for estimating scaling
        coefficients. Default: 0 (x)

    Returns
    -------
    snr: numpy.ndarray
        signal-to-noise (variance) ratio [dB]
    err_std: numpy.ndarray
        **SCALED** error standard deviation
    beta: numpy.ndarray
         scaling coefficients (i_scaled = i * beta_i)

    Notes
    -----

    This function estimates the triple collocation errors, the scaling
    parameter :math:`\\beta` and the signal to noise ratio directly from the
    covariances of the dataset. For a general overview and how this function
    and :py:func:`pytesmo.metrics.tcol_error` are related please see
    [Gruber2015]_.

    Estimation of the error variances from the covariances of the datasets
    (e.g. :math:`\\sigma_{XY}` for the covariance between :math:`x` and
    :math:`y`) is done using the following formula:

    .. math::

       \\sigma_{\\varepsilon_x}^2 =
           \\sigma_{X}^2 - \\frac{\\sigma_{XY}\\sigma_{XZ}}{\\sigma_{YZ}}
    .. math::

       \\sigma_{\\varepsilon_y}^2 =
           \\sigma_{Y}^2 - \\frac{\\sigma_{YX}\\sigma_{YZ}}{\\sigma_{XZ}}
    .. math::

       \\sigma_{\\varepsilon_z}^2 =
           \\sigma_{Z}^2 - \\frac{\\sigma_{ZY}\\sigma_{ZX}}{\\sigma_{YX}}

    :math:`\\beta` can also be estimated from the covariances:

    .. math:: \\beta_x = 1
    .. math:: \\beta_y = \\frac{\\sigma_{XZ}}{\\sigma_{YZ}}
    .. math:: \\beta_z=\\frac{\\sigma_{XY}}{\\sigma_{ZY}}

    The signal to noise ratio (SNR) is also calculated from the variances
    and covariances:

    .. math::

       \\text{SNR}_X[dB] = -10\\log\\left(\\frac{\\sigma_{X}^2\\sigma_{YZ}}
                                         {\\sigma_{XY}\\sigma_{XZ}}-1\\right)
    .. math::

       \\text{SNR}_Y[dB] = -10\\log\\left(\\frac{\\sigma_{Y}^2\\sigma_{XZ}}
                                         {\\sigma_{YX}\\sigma_{YZ}}-1\\right)
    .. math::

       \\text{SNR}_Z[dB] = -10\\log\\left(\\frac{\\sigma_{Z}^2\\sigma_{XY}}
                                         {\\sigma_{ZX}\\sigma_{ZY}}-1\\right)

    It is given in dB to make it symmetric around zero. If the value is zero
    it means that the signal variance and the noise variance are equal. +3dB
    means that the signal variance is twice as high as the noise variance.

    References
    ----------
    .. [Gruber2015] Gruber, A., Su, C., Zwieback, S., Crow, W., Dorigo, W.,
       Wagner, W.  (2015). Recent advances in (soil moisture) triple
       collocation analysis.  International Journal of Applied Earth
       Observation and Geoinformation, in review
    """

    cov = np.cov(np.vstack((x, y, z)))

    ind = (0, 1, 2, 0, 1, 2)
    no_ref_ind = np.where(np.arange(3) != ref_ind)[0]

    snr = 10 * np.log10(
        [
            (
                (cov[i, i] * cov[ind[i + 1], ind[i + 2]])
                / (cov[i, ind[i + 1]] * cov[i, ind[i + 2]])
                - 1
            )
            ** (-1)
            for i in np.arange(3)
        ]
    )
    err_var = np.array(
        [
            cov[i, i]
            - (cov[i, ind[i + 1]] * cov[i, ind[i + 2]])
            / cov[ind[i + 1], ind[i + 2]]
            for i in np.arange(3)
        ]
    )

    beta = np.array(
        [
            cov[ref_ind, no_ref_ind[no_ref_ind != i][0]]
            / cov[i, no_ref_ind[no_ref_ind != i][0]]
            if i != ref_ind
            else 1
            for i in np.arange(3)
        ]
    )

    return snr, np.sqrt(err_var) * beta, beta


@deprecated()
def mse(x, y, ddof=0):
    """
    DEPRECATED: use `msd`, `msd_corr`, `msd_bias`, or `msd_var` for the
    individual components or `pytesmo.metrics.msd_decomposition` for the full
    decomposition instead.

    Mean square error (MSE) as a decomposition of the RMSD into individual
    error components. The MSE is the second moment (about the origin) of the
    error, and thus incorporates both the variance of the estimator and
    its bias. For an unbiased estimator, the MSE is the variance of the
    estimator. Like the variance, MSE has the same units of measurement as
    the square of the quantity being estimated.
    The delta degree of freedom keyword (ddof) can be used to correct for
    the case the true variance is unknown and estimated from the population.
    Concretely, the naive sample variance estimator sums the squared deviations
    and divides by n, which is biased. Dividing instead by n - 1 yields an
    unbiased estimator.
    Parameters
    ----------
    x : numpy.ndarray
        First input vector.
    y : numpy.ndarray
        Second input vector.
    ddof : int, optional
        Delta degree of freedom.The divisor used in calculations is N - ddof,
        where N represents the number of elements. By default ddof is zero.
    Returns
    -------
    mse : float
        Mean square error (MSE).
    mse_corr : float
        Correlation component of MSE.
    mse_bias : float
        Bias component of the MSE.
    mse_var : float
        Variance component of the MSE.
    """
    mse_corr = (
        2
        * np.std(x, ddof=ddof)
        * np.std(y, ddof=ddof)
        * (1 - pearsonr(x, y)[0])
    )
    mse_bias = bias(x, y) ** 2
    mse_var = (np.std(x, ddof=ddof) - np.std(y, ddof=ddof)) ** 2
    mse = mse_corr + mse_bias + mse_var
    return mse, mse_corr, mse_bias, mse_var


@deprecated()
def pearsonr(x, y):
    """
    DEPRECATED: use :func:`pytesmo.metrics.pearson_r` instead if you want
    only the value or the confidence interval, and
    :func:`scipy.stats.pearsonr` if you want the p-value.

    Wrapper for scipy.stats.pearsonr. Calculates a Pearson correlation
    coefficient and the p-value for testing non-correlation.

    Parameters
    ----------
    x : numpy.ndarray
        First input vector.
    y : numpy.ndarray
        Second input vector.
    Returns
    -------
    r : float
        Pearson's correlation coefficent.
    p-value : float
        2 tailed p-value.
    See Also
    --------
    scipy.stats.pearsonr
    """
    return stats.pearsonr(x, y)


@np.errstate(invalid="ignore")
@deprecated()
def pearsonr_recursive(
    x, y, n_old=0, sum_xi_yi=0, sum_xi=0, sum_yi=0, sum_x2=0, sum_y2=0
):
    """
    DEPRECATED: use :func:`pytesmo.metrics.pearson_r` instead if you want
    only the value or the confidence interval, and
    :func:`scipy.stats.pearsonr` if you want the p-value.

    Calculate pearson correlation in a recursive manner based on
    .. math::

       r_{xy} = \\frac{n\\sum x_iy_i-\\sum x_i\\sum y_i}
                      {\\sqrt{n\\sum x_i^2-(\\sum x_i)^2}~
                       \\sqrt{n\\sum y_i^2-(\\sum y_i)^2}}
    Parameters
    ----------
    x: numpy.ndarray
        New values for x
    y: numpy.ndarray
        New values for y
    n_old: float, optional
        number of observations from previous pass
    sum_xi_yi: float, optional
        .. math:: \\sum x_iy_i
        from previous pass
    sum_xi: float, optional
        .. math:: \\sum x_i
        from previous pass
    sum_yi: float, optional
        .. math:: \\sum y_i
        from previous pass
    sum_x2: float, optional
        .. math:: \\sum x_i^2
        from previous pass
    sum_y2: float, optional
        .. math:: \\sum y_i^2
        from previous pass
    Returns
    -------
    r: float
        Pearson correlation coefficient
    params: tuple
       tuple of (n_new, sum_xi_yi, sum_xi, sum_yi, sum_x2, sum_y2) .
       Can be used when calling the next iteration as ``*params``.
    """
    n_new = n_old + len(x)
    sum_xi_yi = sum_xi_yi + np.sum(np.multiply(x, y))
    sum_xi = sum_xi + np.sum(x)
    sum_yi = sum_yi + np.sum(y)
    sum_x2 = sum_x2 + np.sum(x ** 2)
    sum_y2 = sum_y2 + np.sum(y ** 2)

    r = (n_new * sum_xi_yi - sum_xi * sum_yi) / (
        np.sqrt(n_new * sum_x2 - sum_xi ** 2)
        * np.sqrt(n_new * sum_y2 - sum_yi ** 2)
    )

    return r, (n_new, sum_xi_yi, sum_xi, sum_yi, sum_x2, sum_y2)


@deprecated()
def pearson_conf(r, n, c=95):
    """
    DEPRECATED: use :func:`pytesmo.metrics.pearson_r` instead.

    Calcalates the confidence interval of a given pearson
    correlation coefficient using a fisher z-transform,
    only valid for correlation coefficients calculated from
    a bivariate normal distribution

    Parameters
    ----------
    r : float or numpy.ndarray
        Correlation coefficient
    n : int or numpy.ndarray
        Number of observations used in determining the correlation coefficient
    c : float
        Level of confidence in percent, from 0-100.
    Returns
    -------
    r_lower : float or numpy.ndarray
        Lower confidence boundary.
    r_upper : float or numpy.ndarray
        Upper confidence boundary.
    """
    # fisher z transform using the arctanh
    z = np.arctanh(r)
    # calculate the standard error
    std_err = 1 / np.sqrt(n - 3)
    # calculate the quantile of the normal distribution
    # for the given confidence level
    n_quant = 1 - (1 - c / 100.0) / 2.0
    norm_z_value = stats.norm.ppf(n_quant)
    # calculate upper and lower limit for normally distributed z
    z_upper = z + std_err * norm_z_value
    z_lower = z - std_err * norm_z_value
    # inverse fisher transform using the tanh
    return np.tanh(z_lower), np.tanh(z_upper)


@deprecated()
def spearmanr(x, y):
    """
    DEPRECATED: use :func:`pytesmo.metrics.spearman_r` instead if you
    want only the value or the confidence interval, and
    :func:`scipy.stats.spearmanr` if you want the p-value.

    Wrapper for scipy.stats.spearmanr. Calculates a Spearman
    rank-order correlation coefficient and the p-value to
    test for non-correlation.

    Parameters
    ----------
    x : numpy.array
        First input vector.
    y : numpy.array
        Second input vector.
    Returns
    -------
    rho : float
        Spearman correlation coefficient
    p-value : float
        The two-sided p-value for a hypothesis test whose null hypothesis
        is that two sets of data are uncorrelated
    See Also
    --------
    scipy.stats.spearmenr
    """
    return stats.spearmanr(x, y)


@deprecated()
def kendalltau(x, y):
    """
    DEPRECATED: use :func:`pytesmo.metrics.kendall_tau` instead if you want
    only the value or the confidence interval, and
    :func:`scipy.stats.kendalltau` if you want the p-value.

    Wrapper for scipy.stats.kendalltau

    Parameters
    ----------
    x : numpy.array
        First input vector.
    y : numpy.array
        Second input vector.
    Returns
    -------
    Kendall's tau : float
        The tau statistic
    p-value : float
        The two-sided p-palue for a hypothesis test whose null hypothesis
        is an absence of association, tau = 0.
    See Also
    --------
    scipy.stats.kendalltau
    """
    return stats.kendalltau(x, y)


def rolling_pr_rmsd(timestamps, data, window_size, center, min_periods):
    """
    DEPRECATED: use the faster version in metrics._fast instead! Only here for
    testing.

    Computation of rolling Pearson R.

    Parameters
    ----------
    timestamps : float64
        Time stamps as julian dates.
    data : numpy.ndarray
        Time series data in 2d array.
    window_size : float
        Window size in fraction of days.
    center : bool
        Set window at the center.
    min_periods : int
        Minimum number of observations in window required for computation.

    Results
    -------
    pr_arr : numpy.array
        Pearson R and p-value.
    """
    pr_arr = np.empty((timestamps.size, 2), dtype=np.float32)
    rmsd_arr = np.empty(timestamps.size, dtype=np.float32)
    ddof = 0

    for i in range(timestamps.size):
        time_diff = timestamps - timestamps[i]

        if center:
            inside_window = np.abs(time_diff) <= window_size
        else:
            inside_window = (time_diff <= 0) & (time_diff > -window_size)

        idx = np.nonzero(inside_window)[0]
        n_obs = inside_window.sum()

        if n_obs == 0 or n_obs < min_periods:
            pr_arr[i, :] = np.nan
        else:
            sub1 = data[idx[0]: idx[-1] + 1, 0]
            sub2 = data[idx[0]: idx[-1] + 1, 1]

            # pearson r
            pr_arr[i, 0] = np.corrcoef(sub1, sub2)[0, 1]

            # p-value
            if np.abs(pr_arr[i, 0]) == 1.0:
                pr_arr[i, 1] = 0.0
            else:
                df = n_obs - 2.0
                t_squared = (
                    pr_arr[i, 0]
                    * pr_arr[i, 0]
                    * (df / ((1.0 - pr_arr[i, 0]) * (1.0 + pr_arr[i, 0])))
                )
                x = df / (df + t_squared)
                x = np.ma.where(x < 1.0, x, 1.0)
                pr_arr[i, 1] = betainc(0.5 * df, 0.5, x)

            # rmsd
            rmsd_arr[i] = np.sqrt(
                np.sum((sub1 - sub2) ** 2) / (sub1.size - ddof)
            )

    return pr_arr, rmsd_arr



# Copyright (c) 2021,Vienna University of Technology,
# Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#   * Neither the name of the Vienna University of Technology, Department of
#     Geodesy and Geoinformation nor the names of its contributors may be used
#     to endorse or promote products derived from this software without specific
#     prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY, DEPARTMENT OF
# GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Test for the data averaging class"""

import warnings

import pandas as pd
import numpy as np
import pytest

from pytesmo.validation_framework.upscaling import Upscaling

with warnings.catch_warnings():
    warnings.filterwarnings("ignore")


@pytest.fixture
def averager():
    manager_parms = {
        "datasets": None,
        "period": None,
        "read_ts_names": None,
    }
    averager = Upscaling(
        ref_class=None,
        others_class={},
        upscaling_lut={},
        manager_parms=manager_parms
    )

    return averager


def test_upscale(averager):
    """Test all upscaling functions"""
    to_upscale = pd.concat(
        [pd.Series(2, index=np.linspace(1,10), name='sm'),
         pd.Series(4, index=np.linspace(1,10), name='sm')],
        axis=1
    )
    # simple check of series averaging
    upscaled = averager.upscale(to_upscale, method="average")
    should = pd.Series(float(3), index=np.linspace(1,10))
    assert upscaled.equals(should)


def test_tstability(averager):
    """Test temporal stability filtering with noisy or uncorrelated series"""
    n_obs = 1000
    points = np.linspace(0, 2*np.pi, n_obs)
    ts = np.sin(points)
    low_corr = np.sin(points + np.pi)
    high_sterr = np.sin(points) + np.random.normal(0, 2, n_obs)
    to_filter = pd.concat(
        [pd.Series(ts, name='sm_1'),
         pd.Series(ts, name='sm_2'),
         pd.Series(ts, name='sm_3'),
         pd.Series(low_corr, name='low_corr'),
         pd.Series(high_sterr, name='high_sterr')],
        axis=1
    )

    filtered = averager.tstability_filter(to_filter)
    should = to_filter.drop(["low_corr", "high_sterr"], axis="columns")
    assert filtered.equals(should)


@pytest.fixture
def series_2_match():
    data_ref = np.arange(30.)
    data2match = data_ref[:-1]
    data2match[2] = np.nan

    ref_ser = pd.Series(
        data_ref,
        index=pd.date_range("2007-01-01 01:00:00", "2007-01-30 01:00:00", freq="D"),
        name="ref"
    ).to_frame()
    match_ser = pd.Series(
        data2match,
        index=pd.date_range("2007-01-01 05:00:00", "2007-01-29 05:00:00", freq="D"),
        name="ref"
    ).to_frame()
    to_match = [ref_ser, match_ser]

    return to_match


def test_temporal_matching(averager, series_2_match):
    """Test temporal matching"""
    matched = averager.temporal_match(series_2_match, drop_missing=False)
    assert len(matched.index) == 30, "Should be matched to the longest timeseries"

    matched = averager.temporal_match(series_2_match, drop_missing=True)
    assert len(matched.index) == 28, "Should drop the row and the missing timestep with a missing value"

    matched = averager.temporal_match(series_2_match, hours=3)
    assert matched.equals(series_2_match[0]), "Should not be matched"


def test_capture_warning(averager, series_2_match):
    def override_read(points, other_name):
        return series_2_match

    averager.lut = {"other_ds": {0 : [0]}}
    averager.datasets = {"other_ds": {"columns": ["ref"]}}
    averager._read = override_read
    res = averager.get_upscaled_ts(
        gpi=0,
        other_name="other_ds",
        **{"hours": 3}
    )
    assert res.equals(series_2_match[0])



"""
This test module runs code from docs/examples ipynbs to make sure that the
documentation is up-to-date with the package, resp. that the notebooks will
build correctly on readthedocs.
"""

import os
from nbconvert.preprocessors import ExecutePreprocessor
import nbformat
import pytest
from pytesmo.utils import rootdir

examples_path = os.path.join(rootdir(), 'docs', 'examples')

@pytest.mark.parametrize("notebook", [
    "anomalies.ipynb",
    "compare_ASCAT_ISMN.ipynb",
    "swi_calc.ipynb",
    "temporal_collocation.ipynb",
    "triple_collocation.ipynb",
    "validation_framework.ipynb",
])
@pytest.mark.skipif(
    not os.path.isdir(examples_path),
    reason=f"Directory '{examples_path}' not found. "
           "Pytesmo is probably not installed in `editable` mode."
)
@pytest.mark.slow
@pytest.mark.doc_example
def test_ipython_notebook(notebook):
    """
    Run selected ipynb example files from docs/examples as tests.
    IMPORTANT: In order that the paths to testdata from notebooks are also
    applicable to the tests here, this file must be within a sub-folder of
    the tests/ directory (assuming that examples are in docs/examples)!
    """
    preprocessor = ExecutePreprocessor(timeout=600, kernel_name="python3")
    with open(os.path.join(examples_path, notebook)) as f:
        nb = nbformat.read(f, as_version=4)
    preprocessor.preprocess(nb)



# Copyright (c) 2013,Vienna University of Technology, Department of Geodesy and
# Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the Vienna University of Technology, Department of
#      Geodesy and Geoinformation nor the names of its contributors may be used
#      to endorse or promote products derived from this software without
#      specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY, DEPARTMENT
# OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Created on Sep 24, 2013

@author: Christoph.Paulik@geo.tuwien.ac.at
"""

import itertools
import pandas as pd

import pytesmo.temporal_matching as temporal_matching


class BasicTemporalMatching(object):
    """
    Temporal matching object

    Parameters
    ----------
    window : float
        window size to use for temporal matching. A match in other will only
        be found if it is +- window size days away from a point in reference
    """

    def __init__(self, window=0.5):
        self.window = window

    def match(self, reference, *args):
        """
        takes reference and other dataframe and returnes a joined Dataframe
        in this case the reference dataset for the grid is also the
        temporal reference dataset
        """
        ref_df = pd.DataFrame(reference)
        return temporal_matching.combined_temporal_collocation(
            ref_df,
            args,
            self.window,
            dropna=True,
            dropduplicates=True,
            add_ref_data=True,
            combined_dropna="all",
        )

    def combinatory_matcher(self, df_dict, refkey, n=2, **kwargs):
        """
        Basic temporal matcher that matches always one Dataframe to
        the reference Dataframe resulting in matched DataFrame pairs.

        If the input dict has the keys 'data1' and 'data2' then the
        output dict will have the key ('data1', 'data2'). The new key
        is stored as a tuple to avoid any issues with string concetanation.

        During matching the column names of the dataframes will be
        transformed into MultiIndex to ensure unique names.

        Parameters
        ----------
        df_dict: dict of pandas.DataFrames
            dictionary containing the spatially colocated DataFrames.
        refkey: string
            key into the df_dict of the DataFrame that should be taken
            as a reference.
        n: int
            number of datasets to match at once
        k : dummy argument

        Returns
        -------
        matched: dict of pandas.DataFrames
            Dictionary containing matched DataFrames. The key is put
            together from the keys of the input dict as a tuple of the
            keys of the datasets this dataframe contains.
        """
        matched = {}
        keys = list(df_dict)
        keys.pop(keys.index(refkey))
        ref_df = df_dict[refkey]
        ref_df = df_name_multiindex(ref_df, refkey)

        for iterkeys in itertools.combinations(keys, n - 1):
            match_list = []
            match_key = []
            for key in iterkeys:
                other_df = df_dict[key]
                other_df = df_name_multiindex(other_df, key)
                match_list.append(other_df)
                match_key.append(key)

            matched_key = tuple([refkey] + sorted(match_key))
            joined = self.match(ref_df, *match_list)

            if len(joined) != 0:
                matched[matched_key] = joined

        return matched


def dfdict_combined_temporal_collocation(
    dfs, refname, k, window=None, n=None, **kwargs
):
    """
    Applies :py:func:`combined_temporal_collocation` on a dictionary of
    dataframes.

    Parameters
    ----------
    dfs : dict
        Dictionary of pd.DataFrames containing the dataframes to be collocated.
    refname : str
        Name of the reference frame in `dfs`.
    k : int
        Number of columns that will be put together in the output dictionary.
        The output will consist of all combinations of size k.
    window : pd.Timedelta or float, optional
        Window around reference timestamps in which to look for data. Floats
        are interpreted as number of days. If it is not given, defaults to 1
        hour to mimick the behaviour of
        ``BasicTemporalMatching.combinatory_matcher``.
    **kwargs :
        Keyword arguments passed to :py:func:`combined_temporal_collocation`.

    Returns:
    --------
    matched_dict : dict
        Dictionary where the key is tuples of ``(refname, othernames...)``.
    """
    if n is not None:
        if len(dfs) != n:
            return {}

    if window is None:
        window = pd.Timedelta(hours=1)

    others = []
    for name in dfs:
        if name != refname:
            others.append(df_name_multiindex(dfs[name], name))
    ref = df_name_multiindex(dfs[refname], refname)
    matched_df = temporal_matching.combined_temporal_collocation(
        ref, others, window, add_ref_data=True, **kwargs
    )

    # unpack again to dictionary
    matched_dict = {}
    othernames = list(dfs.keys())
    othernames.remove(refname)
    key = tuple([refname] + othernames)
    matched_dict[key] = matched_df
    return matched_dict


def make_combined_temporal_matcher(window):
    """
    Matches multiple dataframes together to only have common timestamps.

    See
    :py:func:`pytesmo.temporal_matching.dfdict_combined_temporal_collocation`
    for more details

    Parameters
    ----------
    window : pd.Timedelta or float, optional
        Window around reference timestamps in which to look for data. Floats
        are interpreted as number of days. If it is not given, defaults to 1
        hour to mimick the behaviour of
        ``BasicTemporalMatching.combinatory_matcher``.
    """

    def matcher(dfs, refname, k=None, **kwargs):
        # this comes from Validation.temporal_match_datasets but is not
        # required
        return dfdict_combined_temporal_collocation(
            dfs,
            refname,
            k,
            window=window,
            dropna=True,
            combined_dropna="any",
            dropduplicates=True,
            **kwargs
        )

    return matcher


def df_name_multiindex(df, name):
    """
    Rename columns of a DataFrame by using new column names that
    are tuples of (name, column_name) to ensure unique column names
    that can also be split again. This transforms the columns to a MultiIndex.
    """
    d = {}
    for c in df.columns:
        d[c] = (name, c)

    return df.rename(columns=d)



# Copyright (c) 2013,Vienna University of Technology,
# Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the Vienna University of Technology,
#      Department of Geodesy and Geoinformation nor the
#      names of its contributors may be used to endorse or promote products
#      derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Module contains wrappers for methods in pytesmo.metrics which can be given
pandas.DataFrames instead of single numpy.arrays.
If the DataFrame has more columns than the function has input parameters
the function will be applied pairwise, resp. to triples.
"""

from collections import namedtuple, OrderedDict
from collections.abc import Iterable
import itertools
import numpy as np
import pandas as pd
from scipy import stats
import warnings

import pytesmo.metrics as metrics
from pytesmo.utils import array_dropna, deprecated


def n_combinations(iterable, n, must_include=None, permutations=False):
    """
    Create possible combinations of an input iterable.

    Parameters
    ---------
    iterable: Iterable
        Elements from this iterable are combined.
    n : int
        Number of elements per combination.
    must_include : Iterable, optional (default: None)
        One or more element(s) of iterable that MUST be in each combination.
    permutations : bool, optional (default: False)
        Create combinations of n elements, order matters: e.g. AB -> AB, BA
        If this is False, the output combinations will be sorted.

    Returns:
    ---------
    combs: iterable
        The possible combinations of n elements.
    """
    if must_include:
        if (not isinstance(must_include, Iterable)) or isinstance(
            must_include, str
        ):
            must_include = [must_include]

    if permutations:
        combs = [c for c in itertools.permutations(iterable, n)]
    else:
        combs = list(itertools.combinations(iterable, n))
    if must_include:
        combs_filtered = []
        for comb in combs:
            if all([i in comb for i in must_include]):
                combs_filtered.append(comb)
        combs = combs_filtered
    return combs


def _wrap_metric(metric, symmetric=True, name=None):
    """
    Wraps a metric function to be called by only providing a dataframe.

    Parameters
    ----------
    metric : callable
        Metric function from pytesmo.metrics
    symmetric : bool, optional
        Whether the metric is symmetrical w.r.t to the order of input
        arguments. Default is ``True``.
    name : str or None, optional
        The name of the namedtuple. If it is None (default), the name of the
        metric will be used.

    Returns
    -------
    wrapped : callable
        New function that takes a dataframe as input and returns the metric
        value(s) as named tuple.
        The name of the function is the same as the input function name.
    """
    metric_name = metric.__name__
    if name is None:
        name = metric_name

    def wrapped(df):
        return _dict_to_namedtuple(
            nwise_apply(df, metric, n=2, comm=symmetric), metric_name
        )

    # add name and docstring
    wrapped.__name__ = name
    wrapped.__doc__ = f"""
    Wrapper to call :py:func:`pytesmo.metrics.{metric_name}` on a dataframe

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe for whose columns combinations the metric should be
        evaluated.

    Returns
    -------
    result : namedtuple
        Metric values for the different combinations. Member names are `df`'s
        column names separated by '_and_'.

    See also :py:func:`pytesmo.metrics.{metric_name}` docstring.
    """
    return wrapped


# add functions that are simple to wrap to the module
bias = _wrap_metric(metrics.bias, symmetric=False)
msd = _wrap_metric(metrics.msd)
rmsd = _wrap_metric(metrics.rmsd)
nrmsd = _wrap_metric(metrics.nrmsd)
ubrmsd = _wrap_metric(metrics.ubrmsd)
mse_corr = _wrap_metric(metrics.mse_corr)
mse_var = _wrap_metric(metrics.mse_var)
mse_bias = _wrap_metric(metrics.mse_bias)
pearson_r = _wrap_metric(metrics.pearson_r)
spearman_r = _wrap_metric(metrics.spearman_r)
kendall_tau = _wrap_metric(metrics.kendall_tau)
nash_sutcliffe = _wrap_metric(metrics.nash_sutcliffe, name="Nash_Sutcliffe")
RSS = _wrap_metric(metrics.RSS)


def mse_decomposition(df):
    """
    Mean square error (MSE) and decomposition of the MSE into individual
    error components.

    Returns
    -------
    result : namedtuple
        with column names of df for which the calculation
        was done as name of the
        element separated by '_and_'

    See Also
    --------
    pytesmo.metrics.mse_decomposition
    """
    mse, mse_corr, mse_bias, mse_var = nwise_apply(
        df, metrics.mse_decomposition, n=2, comm=True
    )
    return (
        _dict_to_namedtuple(mse, "MSE"),
        _dict_to_namedtuple(mse_corr, "MSEcorr"),
        _dict_to_namedtuple(mse_bias, "MSEbias"),
        _dict_to_namedtuple(mse_var, "MSEvar"),
    )


@deprecated()
def mse(df):
    """
    Deprecated: use :py:func:`pytesmo.df_metrics.msd` and the functions for the
    individual components instead, or
    :py:func:`pytesmo.df_metrics.msd_decomposition` for the old functionality
    with better performance.

    Mean square error (MSE) as a decomposition of the RMSD into individual
    error components

    Returns
    -------
    result : namedtuple
        with column names of df for which the calculation
        was done as name of the
        element separated by '_and_'

    See Also
    --------
    pytesmo.metrics.mse

    """
    MSE, MSEcorr, MSEbias, MSEvar = nwise_apply(
        df, metrics.mse, n=2, comm=True
    )
    return (
        _dict_to_namedtuple(MSE, "MSE"),
        _dict_to_namedtuple(MSEcorr, "MSEcorr"),
        _dict_to_namedtuple(MSEbias, "MSEbias"),
        _dict_to_namedtuple(MSEvar, "MSEvar"),
    )


@deprecated()
def tcol_error(df):
    """
    Deprecated: use :py:func:`pytesmo.df_metrics.tcol_metrics` instead.

    Triple collocation error estimate, applied to triples of columns of the
    passed data frame.

    Returns
    -------
    triple_collocation_error_x : namedtuple
        Error for the first dataset
    triple_collocation_error_y : namedtuple
        Error for the second dataset
    triple_collocation_error_z : namedtuple
        Error for the third dataset

    See Also
    --------
    pytesmo.metrics.tcol_error
    """
    # For TC, the input order has NO effect --> comm=True
    err0, err1, err2 = nwise_apply(df, metrics.tcol_error, n=3, comm=True)
    trips = list(err0.keys())  # triples in all err are equal
    assert trips == list(err0.keys()) == list(err1.keys()) == list(err2.keys())

    errors = []
    for trip in trips:
        res = [err0[trip], err1[trip], err2[trip]]
        Inner = namedtuple(
            "triple_collocation_error", OrderedDict(zip(trip, res))
        )
        errors.append(Inner(*res))

    return tuple(errors)


@deprecated()
def tcol_snr(df, ref_ind=0):
    """DEPRECATED: use `tcol_metrics` instead."""
    return tcol_metrics(df, ref_ind=0)


def tcol_metrics(df, ref_ind=0):
    """
    Triple Collocation metrics applied to triples of dataframe columns.

    Parameters
    ----------
    df : pd.DataFrame
        Contains the input values as time series in the df columns
    ref_ind : int or None, optional (default: 0)
        The index of the column in df that contains the reference data set.
        If None is passed, we use the first column of each triple as the
        reference, otherwise only triples that contain the reference
        dataset are considered during processing.

    Returns
    -------
    snr : namedtuple
        signal-to-noise (variance) ratio [dB] from the named columns.
    err_std_dev : namedtuple
        **SCALED** error standard deviation from the named columns
    beta : namedtuple
        Scaling coefficients (i_scaled = i * beta_i)
    """
    # For TC, the input order has NO effect --> comm=True
    if ref_ind is not None:
        # This column must be part of each triple and is always used as the
        # reference
        incl = [ref_ind]
    else:
        # All unique triples are processed, the first dataset of a triple is
        # the reference.
        incl = None
        ref_ind = 0
    snr, err, beta = nwise_apply(
        df,
        metrics.tcol_metrics,
        n=3,
        comm=True,
        must_include=incl,
        ref_ind=ref_ind,
    )

    results = {}
    var_dict = {"snr": snr, "err_std_dev": err, "beta": beta}
    for var_name, var_vals in var_dict.items():
        results[var_name] = []
        for trip, res in var_vals.items():
            Inner = namedtuple(var_name, OrderedDict(zip(trip, res)))
            results[var_name].append(Inner(*res))

    return (results["snr"], results["err_std_dev"], results["beta"])


def pearsonr(df):
    """
    Wrapper for scipy.stats.pearsonr

    Returns
    -------
    result : namedtuple
        with column names of df for which the calculation
        was done as name of the
        element separated by '_and_'

    See Also
    --------
    pytesmo.metrics.pearsonr
    scipy.stats.pearsonr
    """
    r, p = nwise_apply(df, stats.pearsonr, n=2, comm=True)
    return (
        _dict_to_namedtuple(r, "Pearsons_r"),
        _dict_to_namedtuple(p, "p_value"),
    )


def spearmanr(df):
    """
    Wrapper for scipy.stats.spearmanr

    Returns
    -------
    result : namedtuple
        with column names of df for which the calculation
        was done as name of the
        element separated by '_and_'

    See Also
    --------
    pytesmo.metrics.spearmenr
    scipy.stats.spearmenr
    """
    r, p = nwise_apply(df, stats.spearmanr, n=2, comm=True)
    return (
        _dict_to_namedtuple(r, "Spearman_r"),
        _dict_to_namedtuple(p, "p_value"),
    )


def kendalltau(df):
    """
    Wrapper for scipy.stats.kendalltau

    Returns
    -------
    result : namedtuple
        with column names of df for which the calculation
        was done as name of the
        element separated by '_and_'

    See Also
    --------
    pytesmo.metrics.kendalltau
    scipy.stats.kendalltau
    """
    r, p = nwise_apply(df, stats.kendalltau, n=2, comm=True)
    return (
        _dict_to_namedtuple(r, "Kendall_tau"),
        _dict_to_namedtuple(p, "p_value"),
    )


def pairwise_apply(df, method, comm=False):
    """
    Compute given method pairwise for all columns, excluding NA/null values

    Parameters
    ----------
    df : pd.DataFrame
        input data, method will be applied to each column pair
    method : function
        method to apply to each column pair. has to take 2 input arguments of
        type np.array and return one value or tuple of values
    comm : bool, optional (default: False)
        Also fills the lower part of the results matrix

    Returns
    -------
    results : pd.DataFrame
    """
    warnings.warn(
        "pairwise_apply() is deprecated, use nwise_apply(..., n=2) instead",
        DeprecationWarning,
    )
    numeric_df = df._get_numeric_data()
    cols = numeric_df.columns
    mat = numeric_df.values
    mat = mat.T
    applyf = method
    K = len(cols)
    result_empty = np.empty((K, K), dtype=float)
    result_empty.fill(np.nan)

    # find out how many variables the applyf returns
    c = applyf(mat[0], mat[0])
    result = []
    for index, value in enumerate(np.atleast_1d(c)):
        result.append(result_empty)
    result = np.array(result)
    mask = np.isfinite(mat)
    for i, ac in enumerate(mat):
        for j, bc in enumerate(mat):
            if i == j:
                continue
            if comm and np.isfinite(result[0][i, j]):
                continue
            valid = mask[i] & mask[j]
            if not valid.any():
                continue
            if not valid.all():
                c = applyf(ac[valid], bc[valid])
            else:
                c = applyf(ac, bc)

            for index, value in enumerate(np.atleast_1d(c)):
                result[index][i, j] = value
                if comm:
                    result[index][j, i] = value
    return_list = []
    for data in result:
        return_list.append(df._constructor(data, index=cols, columns=cols))

    if len(return_list) == 1:
        return return_list[0]
    else:
        return tuple(return_list)


def nwise_apply(
    df,
    method,
    n=2,
    comm=False,
    as_df=False,
    ds_names=True,
    must_include=None,
    **method_kwargs,
):
    """
    Compute given method for column combinations of a data frame, excluding
    NA/null values.

    Parameters
    ----------
    df : pd.DataFrame
        Input data, method will be applied to combinations of columns of this
        df.
    method : function
        method to apply to each column pair. Has to take 2 input arguments of
        type numpy.array and return one value or tuple of values
    n : int, optional (default: 2)
        Number of columns that are combined. The default n=2 is the same as the
        previous pairwise_apply() function.
    comm : bool, optional (default: False)
        Metrics do NOT depend on the order of input values. In these cases we
        can skip unnecessary calculations and simply copy the results if
        necessary (faster).
    as_df : bool, optional (default: False)
        Return matrix structure, same as for previous pairwise_apply(), only
        available for n=2. By default, the return value will be a list of
        ordered dicts.
    ds_names : bool, optional (default: True)
        Use the column names of df to identify the dataset instead of using
        their index.
    must_include : list, optional (default: None)
        The index of one or multiple columns in df that MUST be in part of each
        combination that is processed.
    method_kwargs :
        Keyword arguments that are passed to method.

    Returns
    -------
    results : pd.DataFrame or dict or tuple
    """

    numeric_df = df._get_numeric_data()
    cols = numeric_df.columns.values
    mat = numeric_df.values
    mat = mat.T
    applyf = method

    mask = np.isfinite(mat)

    # create the possible combinations of lines
    counter = list(range(mat.shape[0]))  # get the number of lines?
    # ALL possible combinations of lines?
    perm = True if not comm else False
    combs = n_combinations(
        counter, n, must_include=must_include, permutations=perm
    )

    # find out how many variables the applyf returns
    result = []
    # apply the method using the first data set to find out the shape of c,
    c = applyf(*array_dropna(*[mat[i] for i in range(n)]))
    for index, value in enumerate(np.atleast_1d(c)):
        result.append(OrderedDict([(c, np.nan) for c in combs]))
    result = np.array(result)  # array of OrderedDicts
    # each return value result is a dict that gets filled with dicts that have
    # the cols and keys and the results as values

    lut_comb_cols = dict()

    for comb in combs:
        valid = np.logical_and(*[mask[i] for i in comb])  # where all are True

        lut_comb_cols.update(dict(zip(comb, tuple(np.take(cols, comb)))))

        if not valid.any():
            continue
        if not valid.all():
            c = applyf(*[mat[i, :][valid] for i in comb], **method_kwargs)
        else:
            c = applyf(*[mat[i, :] for i in comb], **method_kwargs)

        for index, value in enumerate(np.atleast_1d(c)):
            result[index][comb] = value

    if as_df:
        if n != 2:
            raise ValueError("Array structure only available for n=2")
        else:
            if not ds_names:
                lut_comb_cols = None
            result = [
                _to_df(r, comm=comm, lut_names=lut_comb_cols) for r in result
            ]
    else:
        if ds_names:
            formatted_results = []
            for r in result:
                formatted = OrderedDict()
                for k, v in r.items():
                    formatted[tuple([lut_comb_cols[i] for i in k])] = v
                formatted_results.append(formatted)
            result = formatted_results

    if len(result) == 1:
        result = result[0]
    else:
        result = tuple(result)

    return result


def _to_df(result, comm=False, lut_names=None):
    """
    Create a 2d results matrix/dataframe from the result dictionary to
    reproduce the output structure of the previous pairwise_apply() function.

    Parameters
    ---------
    result : OrderedDict
        The results as the are calculated in nwise_apply()
    comm : bool, optional (default: False)
        Copy elements from the upper diagonal matrix in the lower diagonal.
    lut_names: dict, optional (default: None)
        A LUT that applies nice names to the columns and lines in the data
        frame, e.g. {1:'ds1', 2:'ds2', 3:'ds3')
    """

    # find out how large the matrix is
    imax = max([max(r) for r in list(result.keys())])
    # create and fill the matrix
    res = np.full((imax + 1, imax + 1), np.nan)
    for k, v in result.items():
        res[k[::-1]] = v
    res = res.transpose()

    if comm:
        i_upper = np.triu_indices(res.shape[0], 1)
        i_lower = np.tril_indices(res.shape[0], -1)
        res[i_lower] = res[i_upper]

    if lut_names is not None:
        res = pd.DataFrame(
            data={lut_names[i]: res[:, i] for i in list(range(max(res.shape)))}
        )
    else:
        res = pd.DataFrame(
            data={i: res[:, i] for i in list(range(max(res.shape)))}
        )
    res.index = res.columns
    return res


def _dict_to_namedtuple(res_dict, name):
    """
    Takes the OrderedDictionary produced by nwise_apply(..., as_df=False) and
    produces named tuples, using the dictionary keys.
    """

    names = []
    values = []

    for k, v in res_dict.items():
        names.append("_and_".join(k))
        values.append(v)

    result = namedtuple(name, names)
    return result._make(values)



# cython: embedsignature=True
# Copyright (c) 2013,Vienna University of Technology, Department of Geodesy and
# Geoinformation All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the Vienna University of Technology, Department of
#      Geodesy and Geoinformation nor the names of its contributors may be used
#      to endorse or promote products derived from this software without
#      specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY, DEPARTMENT
# OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
Created on Oct 16, 2013

Fast cython functions for calculating various filters

@author: Christoph Paulik christoph.paulik@geo.tuwien.ac.at
'''

import platform

cimport numpy as np
import numpy as np
from libc.math cimport exp, fabs, isnan
cimport cython
from cython cimport floating


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def exp_filter(floating [:] in_data,
               floating [:] in_jd,
               int ctime=10,
               double nan=-999999.0):
    """
    Calculates exponentially smoothed time series using an
    iterative algorithm

    Parameters
    ----------
    in_data : double numpy.array
        input data
    in_jd : double numpy.array
        julian dates of input data
    ctime : int
        characteristic time used for calculating
        the weight
    nan : double
        nan values to exclude from calculation
    """
    cdef np.ndarray[floating, ndim = 1] filtered = np.empty(len(in_data))
    cdef double tdiff
    cdef float ef
    cdef float gain = 1
    cdef bint is_nan
    cdef double last_jd_var
    cdef double last_filtered_var
    cdef unsigned int i
    cdef int found_index = -1

    filtered.fill(np.nan)

    # find the first non nan value in the time series

    for i in range(in_jd.shape[0]):

        is_nan = (in_data[i] == nan) or isnan(in_data[i])
        if not is_nan:
            last_jd_var = in_jd[i]
            last_filtered_var = in_data[i]
            # set the first filtered value to the first found non nan value
            filtered[i] = in_data[i]
            found_index = i
            break

    if found_index > -1:

        for index in range(found_index + 1, in_jd.shape[0]):
            is_nan = (in_data[index] == nan) or isnan(in_data[index])
            if not is_nan:
                tdiff = in_jd[index] - last_jd_var
                ef = exp(-tdiff / ctime)
                gain = gain / (gain + ef)
                filtered[index] = (
                    last_filtered_var
                    + gain * (in_data[index] - last_filtered_var)
                )
                last_jd_var = in_jd[index]
                last_filtered_var = filtered[index]

    return filtered


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def boxcar_filter(floating [:] in_data,
                  floating [:] in_jd,
                  float window=1, double nan=-999999.0,
                  bint fillna=0, int min_obs=1):
    """
    Calculates filtered time series using
    a boxcar filter - basically a moving average calculation

    Parameters
    ----------
    in_data : double numpy.array
        input data
    in_jd : double numpy.array
        julian dates of input data
    window : int
        characteristic time used for calculating
        the weight
    nan : double
        nan values to exclude from calculation
    """
    cdef np.ndarray[floating, ndim = 1] filtered = np.empty(len(in_data))
    cdef double tdiff
    cdef unsigned int i
    cdef unsigned int j
    cdef bint is_nan
    cdef double sum = 0
    cdef double limit = window / 2.0
    cdef int nobs = 0

    filtered.fill(np.nan)

    for i in range(in_jd.shape[0]):
        is_nan = (in_data[i] == nan) or isnan(in_data[i])
        if (not is_nan) or fillna:
            sum = 0
            nobs = 0
            for j in range(i, in_jd.shape[0]):
                is_nan = (in_data[j] == nan) or isnan(in_data[j])
                if not is_nan:
                    tdiff = in_jd[j] - in_jd[i]
                    if fabs(tdiff) <= limit:
                        sum = sum + in_data[j]
                        nobs = nobs + 1
                    else:
                        break

            for j in range(i, -1, -1):
                if i != j:
                    is_nan = (in_data[j] == nan) or isnan(in_data[j])
                    if not is_nan:
                        tdiff = in_jd[j] - in_jd[i]
                        if fabs(tdiff) <= limit:
                            sum = sum + in_data[j]
                            nobs = nobs + 1
                        else:
                            break

            if nobs >= min_obs:
                filtered[i] = sum / nobs

    return filtered



# Copyright (c) 2016,Vienna University of Technology,
# Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#   * Neither the name of the Vienna University of Technology, Department of
#     Geodesy and Geoinformation nor the names of its contributors may be used
#     to endorse or promote products derived from this software without
#     specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY, DEPARTMENT
# OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
Test for temporal matchers
'''

import numpy as np
import pandas as pd
from pathlib import Path

import pytesmo.validation_framework.temporal_matchers as temporal_matchers


def test_combinatory_matcher_n2():

    n = 1000
    x = np.arange(n)
    y = np.arange(n) * 0.5
    index = pd.date_range(start="2000-01-01", periods=n, freq="D")

    df = pd.DataFrame({'x': x, 'y': y}, columns=['x', 'y'], index=index)
    df2 = pd.DataFrame({'x': x, 'y': y}, columns=['x', 'y'], index=index)
    df3 = pd.DataFrame({'x': x, 'y': y}, columns=['x', 'y'], index=index)

    df_dict = {'data1': df,
               'data2': df2,
               'data3': df3}

    temp_matcher = temporal_matchers.BasicTemporalMatching()
    matched = temp_matcher.combinatory_matcher(df_dict, 'data1')
    assert sorted(list(matched)) == sorted([('data1', 'data2'),
                                            ('data1', 'data3')])
    assert sorted(list(matched[('data1',
                                'data2')].columns)) == sorted([('data1', 'x'),
                                                               ('data1', 'y'),
                                                               ('data2', 'x'),
                                                               ('data2', 'y')])

    assert sorted(list(matched[('data1',
                                'data3')].columns)) == sorted([('data1', 'x'),
                                                               ('data1', 'y'),
                                                               ('data3', 'x'),
                                                               ('data3', 'y')])


def test_combinatory_matcher_n3():

    n = 1000
    x = np.arange(n)
    y = np.arange(n) * 0.5
    index = pd.date_range(start="2000-01-01", periods=n, freq="D")

    df = pd.DataFrame({'x': x, 'y': y}, columns=['x', 'y'], index=index)
    df2 = pd.DataFrame({'x': x, 'y': y}, columns=['x', 'y'], index=index)
    df3 = pd.DataFrame({'x': x, 'y': y}, columns=['x', 'y'], index=index)
    df4 = pd.DataFrame({'x': x, 'y': y}, columns=['x', 'y'], index=index)

    df_dict = {'data1': df,
               'data2': df2,
               'data3': df3}

    temp_matcher = temporal_matchers.BasicTemporalMatching()
    matched = temp_matcher.combinatory_matcher(df_dict, 'data1', n=3)
    assert list(matched) == [('data1', 'data2', 'data3')]
    assert sorted(list(matched[('data1',
                                'data2',
                                'data3')].columns)) == sorted([('data1', 'x'),
                                                               ('data1', 'y'),
                                                               ('data2', 'x'),
                                                               ('data2', 'y'),
                                                               ('data3', 'x'),
                                                               ('data3', 'y')])

    df_dict = {'data1': df,
               'data2': df2,
               'data3': df3,
               'data4': df4}

    temp_matcher = temporal_matchers.BasicTemporalMatching()
    matched = temp_matcher.combinatory_matcher(df_dict, 'data1', n=3)
    assert sorted(list(matched)) == sorted([('data1', 'data2', 'data3'),
                                            ('data1', 'data2', 'data4'),
                                            ('data1', 'data3', 'data4')])
    assert sorted(list(matched[('data1',
                                'data2',
                                'data3')].columns)) == sorted([('data1', 'x'),
                                                               ('data1', 'y'),
                                                               ('data2', 'x'),
                                                               ('data2', 'y'),
                                                               ('data3', 'x'),
                                                               ('data3', 'y')])


def test_add_name_to_df_columns():

    n = 10
    x = np.arange(n)
    y = np.arange(n) * 0.5
    index = pd.date_range(start="2000-01-01", periods=n, freq="D")

    df = pd.DataFrame({'x': x, 'y': y}, columns=['x', 'y'], index=index)
    df = temporal_matchers.df_name_multiindex(df, 'test')
    assert list(df.columns) == [('test', 'x'), ('test', 'y')]


def test_dfdict_combined_temporal_collocation():

    ref_dr = pd.date_range("2000", "2020", freq="YS")
    dr1 = pd.date_range("2000", "2015", freq="YS")
    dr2 = pd.date_range("2005", "2020", freq="YS")

    ref_df = pd.DataFrame({"ref": np.arange(len(ref_dr))}, index=ref_dr)
    df1 = pd.DataFrame(
        {"k1": np.arange(len(dr1)), "k2": np.arange(len(dr1))}, index=dr1
    )
    df2 = pd.DataFrame(
        {"k1": np.arange(len(dr2)), "k2": np.arange(len(dr2))}, index=dr2
    )

    dfs = {"refkey": ref_df, "df1key": df1, "df2key": df2}
    window = pd.Timedelta(days=300)

    matched = temporal_matchers.dfdict_combined_temporal_collocation(
        dfs, "refkey", 2, window=window, n=3, combined_dropna=True
    )

    # keys are the same, only refkey is missing
    key = ("refkey", "df1key", "df2key")
    assert list(matched.keys()) == [key]

    # overlap is only 11 timestamps
    assert matched[key].shape == (11, 5)

    overlap_dr = pd.date_range("2005", "2015", freq="YS")
    assert np.all(matched[key].index == overlap_dr)

    # test with ASCAT and ISMN data
    here = Path(__file__).resolve().parent
    ascat = pd.read_csv(here / "ASCAT.csv", index_col=0, parse_dates=True)
    ismn = pd.read_csv(here / "ISMN.csv", index_col=0, parse_dates=True)

    dfs = {"ASCAT": ascat[["sm"]], "ISMN": ismn[["soil_moisture"]]}
    refname = "ISMN"
    window = pd.Timedelta(12, "H")

    old_matcher = temporal_matchers.BasicTemporalMatching().combinatory_matcher
    new_matcher = temporal_matchers.make_combined_temporal_matcher(window)

    expected = old_matcher(dfs, refname, k=2, n=2)
    new = new_matcher(dfs, refname, k=2, n=2)

    key = ("ISMN", "ASCAT")
    assert list(expected.keys()) == [key]
    assert list(new.keys()) == [key]
    # We have to do an extra dropna for the old matcher, because the old
    # matcher doesn't do this by itself.
    # This is normally done within validation.py, `get_data_for_result_tuple`,
    # but since the combined matcher should exclude all data where even a
    # single entry misses (so that all only have common data) this is done
    # before in the new matcher (the combined matcher, whereas the old one is
    # the combinatory matcher)
    exp = expected[key].dropna()
    assert exp.shape == new[key].shape
    for col in new[key]:
        np.testing.assert_equal(exp[col].values, new[key][col].values)



# -*- coding: utf-8 -*-
# <nbformat>3.0</nbformat>

# <headingcell level=1>

# Reading H-SAF images

# <rawcell>

# In this Example we will read and plot images of the H-SAF (soil moisture) products H07, H08 and H14.

# <codecell>

import ascat.h_saf as h_saf
import os
import datetime
import matplotlib.pyplot as plt
from mpl_toolkits.basemap import Basemap
import pytesmo.colormaps.load_cmap as smcolormaps
import pytesmo.grid.resample as resample
import numpy as np
#
#set the paths to the image files
#I'm using the test images included in the pytesmo package

testdata_folder = '/pytesmo/testdata'

h_saf_path = os.path.join(testdata_folder, 'sat/h_saf')
h07_path = os.path.join(h_saf_path, 'h07')
h08_path = os.path.join(h_saf_path, 'h08')
h14_path = os.path.join(h_saf_path, 'h14')

#initialize the readers with the path
h07_reader = h_saf.H07img(h07_path)
h08_reader = h_saf.H08img(h08_path)
h14_reader = h_saf.H14img(h14_path)

# <headingcell level=1>

# Reading the H07 product

# <rawcell>

# pytesmo includes one h07 image with the timestamp 2010-05-01 08:33:01
# We can either read this image alone if we know the timestamp or iterate over all images on 2010-05-01.

# <codecell>

#the reader returns not only the data but also metadata and the longitudes and latitudes
h07_data, metadata, timestamp, lons, lats, time_var = h07_reader.read(datetime.datetime(2010,5,1,8,33,1))

# <rawcell>

# let's inspect the data

# <codecell>

print(type(h07_data))
# the data is a dictionary, each dictionary key contains the array of one variable
print("The following variables are in this image", h07_data.keys())

# <rawcell>

# let's inspect surface soil moisture (ssm)

# <codecell>

print(h07_data['Surface Soil Moisture (Ms)'].shape)
# it is only a 1D array to plot the data we also need latitude and logitude information
print(lons.shape)
print(lats.shape)

# <rawcell>

# let's resample and plot the ssm data
# H07 data is not that easy to plot because it comes in orbit geometry and not on a fixed grid.

# <codecell>

# %matplotlib inline
# lets resample to a 0.1 degree grid
#define the grid points in latitude and logitude
lats_dim = np.arange(25,75,0.1)
lons_dim = np.arange(-25,45,0.1)
#make 2d grid out the 1D grid spacings
lons_grid, lats_grid = np.meshgrid(lons_dim,lats_dim)

resampled_data = resample.resample_to_grid(h07_data, lons, lats,
                                           lons_grid, lats_grid)

fig = plt.figure(figsize=(10,10))
ax = fig.add_axes([0.1,0.1,0.8,0.8])
# setup of basemap for europe
# simple mercator projection
m = Basemap(llcrnrlon=-25.0,llcrnrlat=25.0,urcrnrlon=45.0,urcrnrlat=75.0,\
            resolution='l',area_thresh=1000.,projection='merc',\
            lat_ts=50.,ax=ax)

# make a pseudocolor plot using the ASCAT SWI colormap
im = m.pcolormesh(lons_grid, lats_grid, resampled_data['Surface Soil Moisture (Ms)'], latlon=True,
         cmap=smcolormaps.load('SWI_ASCAT'))

m.drawcoastlines()
m.drawcountries()
# draw parallels and meridians.
# label on left and bottom of map.
parallels = np.arange(20,80,10.)
m.drawparallels(parallels,labels=[1,0,0,0])
meridians = np.arange(-60,50,10.)
m.drawmeridians(meridians,labels=[0,0,0,1])
# add colorbar
cb = m.colorbar(im,"right", size="5%", pad='2%')
ax.set_title('H07 Soil Moisture in %')
plt.show()
plt.close()
fig.clf()
#delete data that is no longer needed to keep memory usage in check
del resampled_data
del h07_data
del lats
del lons
# <headingcell level=1>

# Reading the H08 product

# <rawcell>

# H08 data has a much higher resolution and comes on a 0.00416 degree grid.
# The sample data included in pytesmo was observed on the same time as the included H07 product.
# Instead of read_img you can also use the daily_images iterator. You just specify a day and it will read all the
# images that are in your folder for this day. This also works for the other H07 and H14 reader.

# <codecell>

#the reader returns not only the data but also metadata and the longitudes and latitudes
for h08_data, metadata, timestamp, lons, lats, time_var in h08_reader.daily_images(datetime.datetime(2010,5,1)):
    # this tells you the exact timestamp of the read image
    print(timestamp.isoformat())

    print(type(h08_data))
    # the data is a dictionary, each dictionary key contains the array of one variable
    print("The following variables are in this image", h08_data.keys())
    print(h08_data['ssm'].shape)
    print(lons.shape)
    print(lats.shape)

# <rawcell>

# In our case only one image is in the folder so the loop exits after this image is read
# The data has higher resolution but it already comes as a 2D image.
# Let's plot it.

# <codecell>

# %matplotlib inline
fig = plt.figure(figsize=(10,10))
ax = fig.add_axes([0.1,0.1,0.8,0.8])
# setup of basemap for europe but zoomed in little bit
# simple mercator projection
m = Basemap(llcrnrlon=10,llcrnrlat=50.0,urcrnrlon=45.0,urcrnrlat=75.0,\
            resolution='l',area_thresh=1000.,projection='merc',\
            lat_ts=50.,ax=ax)

# make a pseudocolor plot using the ASCAT SWI colormap
# latitudes and data have to be flipped upside down because the latitudes
# have to be in ascending order to be plotted correctly
# mask values > 100 so that they are not plotted
im = m.pcolormesh(lons, np.flipud(lats), np.ma.masked_greater(np.flipud(h08_data['ssm']),100), latlon=True,
         vmin=0, vmax=100,cmap=smcolormaps.load('SWI_ASCAT'))

m.drawcoastlines()
m.drawcountries()
# draw parallels and meridians.
# label on left and bottom of map.
parallels = np.arange(20,80,10.)
m.drawparallels(parallels,labels=[1,0,0,0])
meridians = np.arange(-60,50,10.)
m.drawmeridians(meridians,labels=[0,0,0,1])
# add colorbar
cb = m.colorbar(im,"right", size="5%", pad='2%')
ax.set_title('H08 Soil Moisture in %')
plt.show()
plt.close()
fig.clf()

#delete data that is no longer needed to keep memory usage in check
del h08_data
del lons
del lats

# <headingcell level=2>

# Reading only area of interest

# <rawcell>

# H08 has a very high resolution, so most people will only want to read it for their area of interest. This can be
# done using the lat_lon_bbox keyword

# <codecell>

#the reader returns not only the data but also metadata and the longitudes and latitudes
h08_roi, metadata, timestamp, lons, lats, time_var = h08_reader.read(datetime.datetime(2010,5,1,8,33,1),
                                                                     lat_lon_bbox=[60,70,15,25])

# %matplotlib inline
fig = plt.figure(figsize=(10,10))
ax = fig.add_axes([0.1,0.1,0.8,0.8])
# setup of basemap for europe but zoomed in little bit
# simple mercator projection
m = Basemap(llcrnrlon=10,llcrnrlat=50.0,urcrnrlon=45.0,urcrnrlat=75.0,\
            resolution='l',area_thresh=1000.,projection='merc',\
            lat_ts=50.,ax=ax)

# make a pseudocolor plot using the ASCAT SWI colormap
# latitudes and data have to be flipped upside down because the latitudes
# have to be in ascending order to be plotted correctly
# mask values > 100 so that they are not plotted
im = m.pcolormesh(lons, np.flipud(lats), np.ma.masked_greater(np.flipud(h08_roi['ssm']),100), latlon=True,
         vmin=0, vmax=100,cmap=smcolormaps.load('SWI_ASCAT'))

m.drawcoastlines()
m.drawcountries()
# draw parallels and meridians.
# label on left and bottom of map.
parallels = np.arange(20,80,10.)
m.drawparallels(parallels,labels=[1,0,0,0])
meridians = np.arange(-60,50,10.)
m.drawmeridians(meridians,labels=[0,0,0,1])
# add colorbar
cb = m.colorbar(im,"right", size="5%", pad='2%')
ax.set_title('H08 Soil Moisture in %')
plt.show()
fig.clf()
plt.close()
#delete data that is no longer needed to keep memory usage in check
del h08_roi
del lats
del lons


# <headingcell level=1>

# Reading the H14 product

# <rawcell>

# The H14 product is a global product on a reduced gaussian grid with a resolution of approx. 25km.

# <codecell>

#the reader returns not only the data but also metadata and the longitudes and latitudes
h14_data, metadata, timestamp, lons, lats, time_var = h14_reader.read(datetime.datetime(2014, 5, 15))

# <codecell>

print(type(h14_data))
# the data is a dictionary, each dictionary key contains the array of one variable
print("The following variables are in this image", h14_data.keys())

# <codecell>

print(h14_data['SM_layer1_0-7cm'].shape)
print(lons.shape)
print(lats.shape)

# <rawcell>

# The data comes as a 2D array. If the keyword expand_grid is set to False during reader initialization then only 1D
# arrays would be returned. This can be good for working with the data but for plotting the expanded grid is easier
# to handle.

# <codecell>

h14_reader_1d = h_saf.H14img(h14_path, expand_grid=False)
#the reader returns not only the data but also metadata and the longitudes and latitudes
h14_data_1d, metadata, timestamp, lons_1d, lats_1d, time_var = h14_reader_1d.read(datetime.datetime(2014, 5, 15))
print(h14_data_1d['SM_layer1_0-7cm'].shape)
print(lons_1d.shape)
print(lats_1d.shape)

# <rawcell>

# Let's plot all layers in the H14 image

# <codecell>

for layer in h14_data:
    # %matplotlib inline
    fig = plt.figure(figsize=(10,10))
    ax = fig.add_axes([0.1,0.1,0.8,0.8])
    # setup of basemap for the world
    # Robinson projection
    m = Basemap(projection='robin',lon_0=0,resolution='c',ax=ax)

    # make a pseudocolor plot using the ASCAT SWI colormap
    # latitudes and data have to be flipped upside down because the latitudes
    # have to be in ascending order to be plotted correctly
    # mask values > 100 so that they are not plotted
    im = m.pcolormesh(lons, lats, h14_data[layer], latlon=True,
             vmin=0, vmax=1,cmap=smcolormaps.load('SWI_ASCAT'))

    m.drawcoastlines()
    m.drawcountries()
    # draw parallels and meridians.
    # label on left and bottom of map.
    parallels = np.arange(-90,90,30.)
    m.drawparallels(parallels,labels=[1,0,0,0])
    meridians = np.arange(-180,180,40.)
    m.drawmeridians(meridians,labels=[0,0,0,1])
    # add colorbar
    cb = m.colorbar(im,"right", size="5%", pad='2%')
    ax.set_title('H14 {:}'.format(layer))
    plt.show()

# <codecell>



# -*- coding: utf-8 -*-

"""
The results manager stores validation results in netcdf format.
"""

from netCDF4 import date2num, num2date, Variable
import numpy as np
import os
import pandas as pd
from pynetcf.base import Dataset
from typing import Union
import warnings


def build_filename(root, key):
    """
    Create savepath/filename that does not exceed 255 characters

    Parameters
    ----------
    root : str
        Directory where the file should be stored
    key : list of tuples
        The keys are joined to create a filename from them. If the length of
        the joined keys is too long we shorten it.

    Returns
    -------
    fname : str
        Full path to the netcdf file to store
    """
    ds_names = []
    for ds in key:
        if isinstance(ds, tuple):
            ds_names.append(".".join(ds))
        else:
            ds_names.append(ds)

    fname = "_with_".join(ds_names)
    ext = "nc"

    if len(os.path.join(root, ".".join([fname, ext]))) > 255:
        ds_names = [str(ds[0]) for ds in key]
        fname = "_with_".join(ds_names)

        if len(os.path.join(root, ".".join([fname, ext]))) > 255:
            fname = "validation"

    return os.path.join(root, ".".join([fname, ext]))


class PointDataResults(Dataset):
    def __init__(self, filename, zlib=True, read_only=False):
        """
        Results manager for validation results as returned by the validation
        framework

        Parameters
        ----------
        filename : str
            Path to the netcdf file to add results to
        zlib : bool, optional (default: True)
            Activate/deactivate file compression
        read_only : bool, optional (default: False)
            Force read only mode when opening the netcdf file
        """

        self.time_unit = "days since 1900-01-01 00:00:00"

        if os.path.exists(filename):
            if read_only:
                mode = "r"
            else:
                mode = "a"
        else:
            mode = "w"

        super(PointDataResults, self).__init__(filename, mode=mode, zlib=zlib)

        if mode == "w":
            # in space (static results)
            self.dataset.createDimension("loc", None)
            # in time (rolling metrics)
            self.dataset.createDimension("obs", None)

            # default variables, along loc dim
            self.write_var(
                "lon",
                dim=("loc",),
                dtype="float",
                attr=dict(
                    long_name="location longitude",
                    standard_name="longitude",
                    units="degrees_east",
                    valid_range=np.array([-180, 180]),
                    axis="X",
                ),
            )

            self.write_var(
                "lat",
                dim=("loc",),
                dtype="float",
                attr=dict(
                    long_name="location latitude",
                    standard_name="latitude",
                    units="degrees_north",
                    valid_range=np.array([-90, 90]),
                    axis="Y",
                ),
            )

            self.write_var(
                "idx",
                dim=("loc",),
                dtype="int",
                attr=dict(
                    long_name="observation point index",
                    standard_name="idx",
                    units="timeseries_id",
                ),
            )

            # indexing var for time series
            self.write_var(
                "_row_size",
                dim=("loc",),
                dtype="int",
                attr=dict(
                    long_name="number of timestamps for this loc",
                    sample_dimension="obs",
                ),
            )

            self.write_var(
                "time",
                dtype="float",
                dim=("obs",),
                attr=dict(
                    long_name="metric time stamp",
                    standard_name="time",
                    units=self.time_unit,
                ),
            )

    def __getitem__(self, key: str) -> Variable:
        return self.dataset[key]

    @property
    def variables(self) -> np.array:
        """Names of all variables in the data set"""
        return self.dataset.variables.keys()

    def _num2date(
        self, nums: np.array, as_pd_idx: bool = True
    ) -> Union[pd.DatetimeIndex, np.array]:
        """Read time stamps and convert them."""
        dates = num2date(
            nums,
            units=self.time_unit,
            calendar="standard",
            only_use_cftime_datetimes=False,
            only_use_python_datetimes=True,
        )

        if as_pd_idx:
            return pd.DatetimeIndex(dates)
        else:
            return dates.astype("datetime64[ns]")

    def _date2num(self, dates: Union[pd.DatetimeIndex, np.array]) -> np.array:
        """Convert datetime index or dates to float"""
        if isinstance(dates, pd.DatetimeIndex):
            dates = dates.to_pydatetime()
        return date2num(dates, units=self.time_unit, calendar="standard")

    def _lonlat2idx(self, lon: float, lat: float) -> int:
        """Find index of location by coordinates"""
        return np.where(
            np.logical_and(self["lon"][:] == lon, self["lat"][:] == lat)
        )[0]

    def _idx2lonlat(self, idx: int) -> tuple:
        """Find coordinates of location by index"""
        return (self["lon"][:][idx], self["lat"][:][idx])

    def _sel_attr(self, attr: dict, name: str, time: bool = False) -> dict:
        """Select attributes for variable"""
        if attr is None:
            attr = {}
        if name in attr.keys():
            var_attr = attr[name]
        else:
            var_attr = {}
        if "coordinates" not in var_attr.keys():
            if time:
                var_attr["coordinates"] = "idx lat lon time"
            else:
                var_attr["coordinates"] = "idx lat lon"

        return var_attr

    def add_ts_results(self, idx, times, results, attr=None):
        """
        Add observations over time to previously added locations results.

        Parameters
        ----------
        idx : int
            Location index, as returned when adding metrics results.
        times : pd.DatetimeIndex or np.array
            Datetime index as in the validation results from rolling metrics
        results : dict
            Variable names as dict keys and data arrays as values.
            Data arrays must have same size as times.
        attr : dict, optional (default: None)
            Variable names as keys and attributes as dicts for each variable.
            Only used when the variable is created, not if it is already in the
            dataset.
        """
        sel = slice(self["_row_size"][:][:idx].sum(), None)
        self["time"][sel] = self._date2num(times)
        self["_row_size"][idx] = len(times)  # update the time series length

        for name, data in results.items():
            if name not in self.variables:
                var_attr = self._sel_attr(attr, name, False)
                self.write_var(
                    name, data=data, dim=("obs",), attr=var_attr
                )
            else:
                self[name][sel] = data

    def add_metrics_results(self, lons, lats, results, attr=None):
        """
        Add observations over time to a locations results.

        Parameters
        ----------
        lons : np.array
            Array of location longitudes, shape must match shape of arrays
            in data
        lats : np.array
            Array of location latitudes, shape must match shape of arrays
            in data
        results : dict
            Variable names as dict keys and data arrays as values. As returned
            by the metric calculators, except the RollingMetrics. Shape of data
            arrays must match lons/lats.
        attr : dict, optional (default: None)
            Variable names as keys and attributes as dicts for each variable.
            Only used when the variable is created, not if it is already in the
            dataset.

        Returns
        -------
        idx : indices
            Indices of the new locations, can be used to add time results
        """
        n = len(lons)
        if not np.all(
            np.array(
                [len(lons), len(lats)] + [len(d) for d in results.values()]
            )
            == n
        ):
            raise ValueError(
                "Lon, Lat and data variable must have the same shape"
            )

        idx = self["idx"][:][-1] + 1 if len(self["idx"]) != 0 else 0
        sel = slice(idx, None)

        self["idx"][sel] = np.arange(idx, idx + n)  # + 1
        self["_row_size"][sel] = 0
        self["lon"][sel], self["lat"][sel] = lons, lats

        for name, data in results.items():
            if isinstance(name, tuple):
                name = "__".join(name)
            if name not in self.variables:
                var_attr = self._sel_attr(attr, name, False)
                self.write_var(name, data=data, dim=("loc",), attr=var_attr)
            else:
                self[name][sel] = data

        return self["idx"][sel].data

    def add_result(self, lon, lat, data, ts_vars=None, times=None, attr=None):
        """
        Add all results (time series and location metrics) for a single point.

        Parameters
        ----------
        lon : float
            Longitude of the point
        lat : float
            Latitude of the point
        data : dict
            Dict of metric names and values. For normal (not rolling) metrics
            this is an array of size 1, otherwise if the same size as time.
        times : np.array, optional (default: None)
            Time values, length must mach all time series (rolling) metrics
            in data.
        attr : dict, optional (default: None)
        """

        if attr is None:
            attr = {}
        if ts_vars is None:
            ts_vars = []

        metric_results = {k: v for k, v in data.items() if k not in ts_vars}
        metric_attrs = {
            k: v for k, v in attr.items() if k in metric_results.keys()
        }

        ts_results = {k: v for k, v in data.items() if k in ts_vars}

        if times is None and ts_results:
            raise ValueError(
                "Got time series variables, "
                "but no times passed: {}".format(ts_results.keys())
            )

        idx = self.add_metrics_results(
            lons=np.array([lon]),
            lats=np.array([lat]),
            results=metric_results,
            attr=metric_attrs,
        )

        if ts_results:
            assert len(idx) == 1
            ts_attrs = {
                k: v for k, v in attr.items() if k in ts_results.keys()
            }
            self.add_ts_results(
                idx[0], times=times, results=ts_results, attr=ts_attrs
            )

    def read_ts(self, idx: int) -> pd.DataFrame:
        """Read time series data for a single point"""
        var_data = {}
        sel = None
        for var in self.variables:
            if var in ["time", "idx", "_row_size"]:
                continue
            if "obs" in self[var].dimensions:
                if sel is None:
                    sel_start = self["_row_size"][:][:idx].sum()
                    sel = slice(
                        sel_start, sel_start + self["_row_size"][:][idx]
                    )
                var_data[var] = self[var][sel]

        return pd.DataFrame(
            index=self._num2date(self["time"][sel]), data=var_data
        )

    def read_loc(self, idx: Union[int, np.array, None] = None) -> pd.DataFrame:
        """Read loc data for one/multiple/all point(s)"""
        if idx is None:
            data = pd.DataFrame(index=self["idx"][:])
            idx = slice(None, None)
        else:
            if isinstance(idx, int):
                idx = [idx]
            data = pd.DataFrame(index=idx)

        for var in self.variables:
            if var in ["time", "idx", "_row_size"]:
                continue
            if "obs" not in self[var].dimensions:
                data[var] = self[var][:][idx]

        return data


def netcdf_results_manager(
    results, save_path, filename: dict = None,
    ts_vars: list = None, zlib=True, attr=None
):
    """
    Write validation results to netcdf file.

    Parameters
    ----------
    results : dict
        Validation results as returned by the metrics calculator.
        Keys are tuples that define the dataset names that were used.
        Values contains 'lon' and 'lat' keys for defining the points, and
        optionally 'time' which sets the time stamps for each location
        (if there are metrics over time in the results - e.g due to
        RollingMetrics)
    save_path : str
        Directory where the netcdf file(s) are are created.
    filename: dict, optional (default: None)
        Filename(s) (value), for each dataset combination in results (key).
        By default (if None is passed) the keys in results are used to
        generate a file name.
    ts_vars : list, optional (default: None)
        List of variables in results that are treated as time series
    zlib : bool, optional (default: True)
        Activate compression
    attr : dict, optional (default: None)
        Variable attributes, variable names as keys, attributes as another
        dict in values.
    """

    if len(results) == 0:
        warnings.warn(f"Empty results, {save_path} will not be created.")
    for ds_names, res in results.items():
        if filename is None:
            fname = build_filename(save_path, ds_names)
        else:
            fname = os.path.join(save_path, filename[ds_names])

        with PointDataResults(fname, zlib=zlib) as writer:
            lons = res.pop("lon")
            lats = res.pop("lat")
            if ts_vars is not None:
                for i, (lon, lat) in enumerate(zip(lons, lats)):
                    data = {}
                    for k, v in res.items():
                        if k.lower() == "time":
                            time = pd.DatetimeIndex(res["time"][i])
                        elif k in ts_vars:
                            data[k] = v[i]
                        else:
                            if not isinstance(v[i], np.ndarray):
                                data[k] = np.array([v[i]])
                            else:
                                data[k] = v[i]

                    writer.add_result(
                        lon,
                        lat,
                        data=data,
                        ts_vars=ts_vars,
                        times=time,
                        attr=attr,
                    )

            else:
                writer.add_metrics_results(lons, lats, results=res, attr=attr)



import numpy as np
import pandas as pd


def create_correlated_data(n_datasets, n, r):
    """Creates n_datasets random timeseries with specified correlation"""
    C = np.ones((n_datasets, n_datasets)) * r
    for i in range(n_datasets):
        C[i, i] = 1
    A = np.linalg.cholesky(C)

    return (A @ np.random.randn(n_datasets, n)).T


class DummyGrid:
    # required for the dummy reader
    def find_nearest_gpi(self, lon, lat, max_dist=np.inf):
        return 0, 0


class DummyReader:

    def __init__(self, dfs, name):
        if not isinstance(dfs, (list, tuple)):
            dfs = [dfs]
        self.data = [pd.DataFrame(dfs[i][name]) for i in range(len(dfs))]
        self.grid = DummyGrid()

    def read(self, gpi, *args, **kwargs):
        return self.data[gpi]


class DummyEmptyDataFrameReader:
    # has data, but only an empty dataframe

    def __init__(self, dfs, name):
        self.data = [pd.DataFrame(dfs[i][name]) for i in range(len(dfs))]
        self.grid = DummyGrid()

    def read(self, gpi, *args, **kwargs):
        names = self.data[gpi].columns
        return pd.DataFrame(np.zeros((0, len(names))), columns=names)


class DummyNoDataReader:
    # has no data at all
    def __init__(self, dfs, name):
        self.grid = DummyGrid()

    def read(self, gpi, *args, **kwargs):
        return []


def create_datasets(n_datasets, npoints, nsamples, missing=False):
    """
    Creates three datasets with given number of points to compare, each
    having number of samples given
    """
    dfs = []
    for gpi in range(npoints):
        r = np.random.rand()
        data = create_correlated_data(n_datasets, nsamples, r)
        index = pd.date_range("1980", periods=nsamples, freq="D")
        dfs.append(
            pd.DataFrame(
                data,
                index=index,
                columns=(["refcol"] +
                         [f"other{i}col" for i in range(1, n_datasets)])))

    datasets = {}
    datasets["0-ERA5"] = {
        "columns": ["refcol"],
        "class": DummyReader(dfs, "refcol")
    }
    for i in range(1, n_datasets - 1):
        datasets[f"{i}-ESA_CCI_SM_combined"] = {
            "columns": [f"other{i}col"],
            "class": DummyReader(dfs, f"other{i}col")
        }
    if missing == "empty":
        datasets[f"{n_datasets - 1}-missing"] = {
            "columns": [f"other{n_datasets - 1}col"],
            "class": DummyEmptyDataFrameReader(dfs, f"other{n_datasets - 1}col")
        }
    elif missing == "nodata":
        datasets[f"{n_datasets - 1}-missing"] = {
            "columns": [f"other{n_datasets - 1}col"],
            "class": DummyNoDataReader(dfs, f"other{n_datasets - 1}col")
        }
    else:
        datasets[f"{n_datasets - 1}-ESA_CCI_SM_combined"] = {
            "columns": [f"other{n_datasets - 1}col"],
            "class": DummyReader(dfs, f"other{n_datasets - 1}col")
        }
    return datasets



import numpy as np
from scipy import stats

from pytesmo.metrics import pairwise
from pytesmo.metrics.tcol import tcol_metrics


def has_analytical_ci(metric_func):
    """
    Checks whether an analytical CI implementation is available.

    Parameters
    ----------
    metric_func : callable
        Function that calculates metric value. Must be from
        :py:mod:`pytesmo.metrics`.

    Returns
    -------
    has_ci : bool
        ``True`` if there is a function with name ``metric_func.__name__ +
        "_ci"`` in :py:mod:`pytesmo.metric_cis`.
    """
    return hasattr(pairwise, metric_func.__name__ + "_ci")


def with_analytical_ci(metric_func, x, y, alpha=0.05):
    """
    Evaluates metric and analytical confidence interval.

    This calculates the metric value and an analytical confidence interval. For
    this to work, the metric function must be from
    :py:mod:`pytesmo.pairwise_metric_cis` Use
    :py:func:`pytesmo.metrics.has_analytical_ci` to check whether it's
    available.

    Parameters
    ----------
    metric_func : callable or str
        Function that calculates metric value. Must be from
        :py:mod:`pytesmo.metrics`, and have an analytical CI function
        implemented in :py:mod:`pytesmo.metric_cis`.
        The metric function must have the following signature:
        ``(x : np.ndarray, y : np.ndarray) -> float``
        Alternatively can be the name of the function in `pytesmo.metrics`.
    x, y : np.ndarray
        Data to be compared.
    alpha : float, optional
        Confidence level, default is 0.05.

    Returns
    -------
    m : float
        Metric value
    lower : float
        Lower bound of confidence interval
    upper : float
        Upper bound of confidence interval

    Raises
    ------
    ValueError :
        If no analytical CI function is available."""
    if isinstance(metric_func, str):
        metric_func = getattr(pairwise, metric_func)
    m = metric_func(x, y)
    name = metric_func.__name__
    if has_analytical_ci(metric_func):
        ci = getattr(pairwise, name + "_ci")(x, y, m, alpha)
    else:
        raise ValueError(f"No analytical CI implemented for {name}.")
    return m, ci[0], ci[1]


def with_bootstrapped_ci(
    metric_func,
    x,
    y,
    alpha=0.05,
    method="percentile",
    nsamples=1000,
    minimum_data_length=100,
):
    """
    Evaluates metric and bootstraps confidence interval.

    This calculates the metric value and uses bootstrapping to find a
    confidence interval for the metric.
    This works only for pairwise metrics, use
    :py:func:`pytesmo.metrics.tcol_metrics_with_bootstrap_ci` for TCA metrics.

    Parameters
    ----------
    metric_func : callable
        Function that calculates metric value.  The metric function must have
        the following signature:
        ``(x : np.ndarray, y : np.ndarray) -> float``
    x, y : np.ndarray
        Data to be compared.
    alpha : float, optional
        Confidence level, default is 0.05.
    method : str, optional
        The method to use to calculate confidence intervals. Available methods
        are:

        - "percentile" (default): Uses the percentiles of the bootstrapped
          metric distribution.
        - "basic": Uses the percentiles of the differences to the original
          metric value.
        - "BCa": Bias-corrected and accelerated bootstrap.

        For more info and a comparison of the methods, see [Gilleland2010]_,
        especially Table 6 on page 38.
    nsamples : int, optional
        Number of bootstrap samples, default is 1000.
    minimum_data_length : int, optional
        Minimum amount of data required to do bootstrapping. Default is 100.

    Returns
    -------
    m : float
        Metric value for pairwise metrics
    lower : float or array of floats
        Lower bound of confidence interval
    upper : float or array of floats
        Upper bound of confidence interval

    References
    ----------
    .. [Gilleland2010] Gilleland, E. (2010). *Confidence Intervals for Forecast\
    Verification* (No. NCAR/TN-479+STR). University Corporation for Atmospheric\
    Research. doi:10.5065/D6WD3XJM
    """
    # Prototype, might be better to implement this in Cython if it's too slow.
    # Then it would probably be best to make this function only a lookup table
    # which calls a cpdef'd method, which itself calls the cdef'd metric
    # function with a cdef'd bootstrap implementation.
    n = len(x)
    if n < minimum_data_length:
        raise ValueError(
            "Not enough data for bootstrapping. Your data should contain at"
            f" least {minimum_data_length} samples.\n"
            f"You can pass 'minimum_data_length={n}' if you want to do"
            "bootstrapping nevertheless."
        )
    orig_metric = metric_func(x, y)
    bootstrapped_metric = np.empty(nsamples, dtype=float)
    if method == "BCa":
        orig_jk = _jackknife(metric_func, x, y)
    for i in range(nsamples):
        idx = np.random.choice(n, size=n)
        _x, _y = x[idx], y[idx]
        bootstrapped_metric[i] = metric_func(_x, _y)
    if method == "percentile":
        lower, upper = _percentile_bs_ci(
            bootstrapped_metric, orig_metric, alpha
        )
    elif method == "basic":
        lower, upper = _basic_bs_ci(bootstrapped_metric, orig_metric, alpha)
    elif method == "BCa":
        lower, upper = _BCa_bs_ci(
            bootstrapped_metric, orig_metric, alpha, orig_jk
        )
    return orig_metric, lower, upper


def tcol_metrics_with_bootstrapped_ci(
    x,
    y,
    z,
    ref_ind=0,
    alpha=0.05,
    method="percentile",
    nsamples=1000,
    minimum_data_length=100,
):
    """
    Evaluates triple collocation metrics and bootstraps confidence interval.

    This calculates the SNR, error standard deviation, and scaling parameter
    value using Triple Collocation Analysis and uses bootstrapping to find
    confidence intervals.

    Parameters
    ----------
    x, y, z : np.ndarray
        Data to be compared.
    ref_ind: int
        Index of reference data set for estimating scaling
        coefficients. Default: 0 (x)
    alpha : float, optional
        Confidence level, default is 0.05.
    method : str, optional
        The method to use to calculate confidence intervals. Available methods
        are:

        - "percentile" (default): Uses the percentiles of the bootstrapped
          metric distribution.
        - "basic": Uses the percentiles of the differences to the original
          metric value.
        - "BCa": Bias-corrected and accelerated bootstrap.

        For more info and a comparison of the methods, see [Gilleland2010]_,
        especially Table 6 on page 38.
    nsamples : int, optional
        Number of bootstrap samples, default is 1000.
    minimum_data_length : int, optional
        Minimum amount of data required to do bootstrapping. Default is 100.

    Returns
    -------
    snr_result : tuple
        ``(snr, lower, upper)``, where each entry is an array of length 3.
    err_std_result : tuple
        ``(err_std, lower, upper)``, where each entry is an array of length 3.
    beta_result : tuple
        ``(beta, lower, upper)``, where each entry is an array of length
        3. Note that beta is always 1 for the reference dataset (see
        `ref_ind`), therefore the lower and upper values are set to 1 too.

    References
    ----------
    .. [Gilleland2010] Gilleland, E. (2010). *Confidence Intervals for Forecast\
    Verification* (No. NCAR/TN-479+STR). University Corporation for Atmospheric\
    Research. doi:10.5065/D6WD3XJM
    """
    # Prototype, might be better to implement this in Cython if it's too slow.
    # Then it would probably be best to make this function only a lookup table
    # which calls a cpdef'd method, which itself calls the cdef'd metric
    # function with a cdef'd bootstrap implementation.
    n = len(x)
    if n < minimum_data_length:
        raise ValueError(
            "Not enough data for bootstrapping. Your data should contain at"
            f" least {minimum_data_length} samples.\n"
            f"You can pass 'minimum_data_length={n}' if you want to do"
            "bootstrapping nevertheless."
        )
    orig_snr, orig_err_std, orig_beta = tcol_metrics(x, y, z, ref_ind=ref_ind)
    bootstrapped_snr = np.empty((nsamples, 3), dtype=float)
    bootstrapped_err_std = np.empty((nsamples, 3), dtype=float)
    bootstrapped_beta = np.empty((nsamples, 3), dtype=float)
    if method == "BCa":
        snr_jk = np.empty((n, 3))
        err_std_jk = np.empty((n, 3))
        beta_jk = np.empty((n, 3))
        mask = np.ones(len(x), dtype=bool)
        for i in range(len(x)):
            mask[i] = False
            snr_jk[i, :], err_std_jk[i, :], beta_jk[i, :] = tcol_metrics(
                x[mask], y[mask], z[mask], ref_ind=ref_ind
            )
            mask[i] = True
    for i in range(nsamples):
        idx = np.random.choice(n, size=n)
        _x, _y, _z = x[idx], y[idx], z[idx]
        (
            bootstrapped_snr[i, :],
            bootstrapped_err_std[i, :],
            bootstrapped_beta[i, :],
        ) = tcol_metrics(_x, _y, _z, ref_ind=ref_ind)

    lower_snr = np.empty(3)
    upper_snr = np.empty(3)
    lower_err_std = np.empty(3)
    upper_err_std = np.empty(3)
    lower_beta = np.empty(3)
    upper_beta = np.empty(3)
    for i in range(3):
        if method == "percentile":
            lower_snr[i], upper_snr[i] = _percentile_bs_ci(
                bootstrapped_snr[:, i], orig_snr[i], alpha
            )
            lower_err_std[i], upper_err_std[i] = _percentile_bs_ci(
                bootstrapped_err_std[:, i], orig_err_std[i], alpha
            )
            if i != ref_ind:
                lower_beta[i], upper_beta[i] = _percentile_bs_ci(
                    bootstrapped_beta[:, i], orig_beta[i], alpha
                )
            else:
                lower_beta[i], upper_beta[i] = 1, 1
        elif method == "basic":
            lower_snr[i], upper_snr[i] = _basic_bs_ci(
                bootstrapped_snr[:, i], orig_snr[i], alpha
            )
            lower_err_std[i], upper_err_std[i] = _basic_bs_ci(
                bootstrapped_err_std[:, i], orig_err_std[i], alpha
            )
            if i != ref_ind:
                lower_beta[i], upper_beta[i] = _basic_bs_ci(
                    bootstrapped_beta[:, i], orig_beta[i], alpha
                )
            else:
                lower_beta[i], upper_beta[i] = 1, 1
        elif method == "BCa":
            lower_snr[i], upper_snr[i] = _BCa_bs_ci(
                bootstrapped_snr[:, i],
                orig_snr[i],
                alpha,
                snr_jk,
            )
            lower_err_std[i], upper_err_std[i] = _BCa_bs_ci(
                bootstrapped_err_std[:, i], orig_err_std[i], alpha, err_std_jk
            )
            if i != ref_ind:
                lower_beta[i], upper_beta[i] = _BCa_bs_ci(
                    bootstrapped_beta[:, i], orig_beta[i], alpha, beta_jk
                )
            else:
                lower_beta[i], upper_beta[i] = 1, 1
    return (
        (orig_snr, lower_snr, upper_snr),
        (orig_err_std, lower_err_std, upper_err_std),
        (orig_beta, lower_beta, upper_beta),
    )


def _jackknife(metric_func, x, y):
    jk = np.empty_like(x)
    mask = np.ones(len(x), dtype=bool)
    for i in range(len(x)):
        mask[i] = False
        jk[i] = metric_func(x[mask], y[mask])
        mask[i] = True
    return jk


def _percentile_bs_ci(bs_m, m, alpha):
    """Calculates the CI using percentiles"""
    lower = np.nanquantile(bs_m, alpha / 2)
    upper = np.nanquantile(bs_m, 1 - alpha / 2)
    return lower, upper


def _basic_bs_ci(bs_m, m, alpha):
    """Basic bootstrap"""
    lower = 2 * m - np.nanquantile(bs_m, 1 - alpha / 2)
    upper = 2 * m - np.nanquantile(bs_m, alpha / 2)
    return lower, upper


def _BCa_bs_ci(bs_m, m, alpha, jk):
    """BCa bootstrap"""
    # see also here regarding implementation:
    # https://www.erikdrysdale.com/bca_python/
    z_alpha = stats.norm.ppf(alpha)
    z_1_alpha = stats.norm.ppf(1 - alpha)

    # bias correction
    z0 = stats.norm.ppf(np.mean(bs_m <= m))

    # acceleration
    jk_mean = np.mean(jk)
    a = np.sum((jk_mean - jk) ** 3) / (
        6 * (np.sum((jk_mean - jk) ** 2)) ** (1.5)
    )

    # calculate adjusted percentiles
    alpha_lower = stats.norm.cdf(
        z0 + (z0 + z_alpha) / (1 - a * (z0 + z_alpha))
    )
    alpha_upper = stats.norm.cdf(
        z0 + (z0 + z_1_alpha) / (1 - a * (z0 + z_1_alpha))
    )
    lower = np.nanquantile(bs_m, alpha_lower)
    upper = np.nanquantile(bs_m, alpha_upper)
    return lower, upper



#Copyright (c) 2013,Vienna University of Technology, Department of Geodesy and Geoinformation
#All rights reserved.

#Redistribution and use in source and binary forms, with or without
#modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the Vienna University of Technology, Department of Geodesy and Geoinformation nor the
#      names of its contributors may be used to endorse or promote products
#      derived from this software without specific prior written permission.

#THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY, 
#DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY
#DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
Example script that reads ASCAT Surface Soil Moisture(SSM) 
and ASCAT Soil Water Index(SWI) for a given latitude and
longitude. It also applies different maskings.


Created on Aug 8, 2013

@author: Christoph Paulik Christoph.Paulik@geo.tuwien.ac.at
'''


from ascat.tuw import Ascat_SWI
from ascat.tuw import Ascat_SSM
from ascat.cgls import SWI_TS
import os
import matplotlib.pyplot as plt


#path to ASCAT SSM zip files from FTP server
#on windows the first string has to be your drive letter
#like 'C:\\'

testdata_folder = '/pytesmo/testdata'

path_to_ascat_ssm_data = os.path.join(testdata_folder,'sat/ascat/SSM') 

path_to_ascat_swi_data = os.path.join(testdata_folder,'sat/ascat/SWI') 

path_to_ascat_swi_ts_data = os.path.join(testdata_folder,'sat/cglops/swi_ts') 


#path to grid definition file, default name TUW_W54_01_lonlat-ld-land.txt
path_to_grid_definition = os.path.join(testdata_folder,'sat/ascat/grid')  

#path to advisory flags from FTP Server
path_to_adv_flags = os.path.join(testdata_folder,'sat/ascat/advisory_flags')   

#init the ASCAT_SSM reader with the paths
ascat_SSM_reader = Ascat_SSM(path_to_ascat_ssm_data,path_to_grid_definition,
                                   advisory_flags_path = path_to_adv_flags)

#lon, lat = 14.284130, 45.698074
lon, lat = 14.284, 45.699

#reads ssm data nearest to this lon,lat coordinates
ssm_data_raw = ascat_SSM_reader.read_ssm(lon,lat)

#plot the data using pandas builtin plot functionality
ssm_data_raw.plot()
plt.show()


#read the same data but mask observations where the SSF shows frozen
#and where frozen and snow probabilty are greater than 20%
ssm_data_masked = ascat_SSM_reader.read_ssm(lon,lat,mask_ssf=True,mask_frozen_prob=20,mask_snow_prob=20)

#plot the data using pandas builtin plot functionality
#this time using a subplot for each variable in the DataFrame
ssm_data_masked.plot(subplots=False)
plt.show()

#plot raw and masked SSM data in one plot to compare them

ssm_data_raw.data['SSM'].plot(label='raw SSM data')
ssm_data_masked.data['SSM'].plot(label='masked SSM data')
plt.legend()
plt.show()


#read SWI data using the SWI_TS reader
ascat_SWI_reader=SWI_TS(path_to_ascat_swi_ts_data)
swi_data_raw = ascat_SWI_reader.read_ts(3002621)

#plot the data using pandas builtin plot functionality
swi_data_raw.plot()
plt.show()



#read SWI data using the Ascat_SWI reader
ascat_SWI_reader = Ascat_SWI(path_to_ascat_swi_data,path_to_grid_definition,advisory_flags_path = path_to_adv_flags)


#reads swi data nearest to this lon,lat coordinates
#without any additional keywords all unmasked T values and
#Quality flags will be read
swi_data_raw = ascat_SWI_reader.read_swi(lon,lat)

#plot the data using pandas builtin plot functionality
swi_data_raw.plot()
plt.show()


#read the same data but this time only SWI with a T value
#of 20 is returned
swi_data_T_20 = ascat_SWI_reader.read_swi(lon,lat,T=20)

#plot the data using pandas builtin plot functionality
#this time using a subplot for each variable in the DataFrame
swi_data_T_20.plot(subplots=True)
plt.show()

#you can also mask manually if you prefer
swi_data_T_20.data = swi_data_T_20.data[swi_data_T_20.data['frozen_prob'] < 10]

swi_data_T_20.plot(subplots=True)
plt.show()















"""
Triple collocation metrics.

To avoid recalculation of the covariance matrix, all metrics are calculated
inside :py:func:`tcol_metrics`. Bootstrapped CIs can be obtained with
:py:func:`tcol_metrics_with_bootstrapped_ci`.

Bootstrapping is currently not available for extended TCA.
"""


from itertools import permutations, combinations
import numpy as np


@np.errstate(invalid="ignore")
def tcol_metrics(x, y, z, ref_ind=0):
    """
    Triple collocation based estimation of signal-to-noise ratio, absolute
    errors, and rescaling coefficients

    Parameters
    ----------
    x: 1D numpy.ndarray
        first input dataset
    y: 1D numpy.ndarray
        second input dataset
    z: 1D numpy.ndarray
        third input dataset
    ref_ind: int
        Index of reference data set for estimating scaling
        coefficients. Default: 0 (x)

    Returns
    -------
    snr: numpy.ndarray
        signal-to-noise (variance) ratio [dB]
    err_std: numpy.ndarray
        **SCALED** error standard deviation
    beta: numpy.ndarray
         scaling coefficients (i_scaled = i * beta_i)

    Notes
    -----

    This function estimates the triple collocation errors, the scaling
    parameter :math:`\\beta` and the signal to noise ratio directly from the
    covariances of the dataset. For a general overview and how this function
    and :py:func:`pytesmo.metrics.tcol_error` are related please see
    [Gruber2015]_.

    Estimation of the error variances from the covariances of the datasets
    (e.g. :math:`\\sigma_{XY}` for the covariance between :math:`x` and
    :math:`y`) is done using the following formula:

    .. math::

       \\sigma_{\\varepsilon_x}^2 =
           \\sigma_{X}^2 - \\frac{\\sigma_{XY}\\sigma_{XZ}}{\\sigma_{YZ}}
    .. math::

       \\sigma_{\\varepsilon_y}^2 =
           \\sigma_{Y}^2 - \\frac{\\sigma_{YX}\\sigma_{YZ}}{\\sigma_{XZ}}
    .. math::

       \\sigma_{\\varepsilon_z}^2 =
           \\sigma_{Z}^2 - \\frac{\\sigma_{ZY}\\sigma_{ZX}}{\\sigma_{YX}}

    :math:`\\beta` can also be estimated from the covariances:

    .. math:: \\beta_x = 1
    .. math:: \\beta_y = \\frac{\\sigma_{XZ}}{\\sigma_{YZ}}
    .. math:: \\beta_z=\\frac{\\sigma_{XY}}{\\sigma_{ZY}}

    The signal to noise ratio (SNR) is also calculated from the variances
    and covariances:

    .. math::

       \\text{SNR}_X[dB] = -10\\log\\left(\\frac{\\sigma_{X}^2\\sigma_{YZ}}
                                         {\\sigma_{XY}\\sigma_{XZ}}-1\\right)
    .. math::

       \\text{SNR}_Y[dB] = -10\\log\\left(\\frac{\\sigma_{Y}^2\\sigma_{XZ}}
                                         {\\sigma_{YX}\\sigma_{YZ}}-1\\right)
    .. math::

       \\text{SNR}_Z[dB] = -10\\log\\left(\\frac{\\sigma_{Z}^2\\sigma_{XY}}
                                         {\\sigma_{ZX}\\sigma_{ZY}}-1\\right)

    It is given in dB to make it symmetric around zero. If the value is zero
    it means that the signal variance and the noise variance are equal. +3dB
    means that the signal variance is twice as high as the noise variance.

    References
    ----------
    .. [Gruber2015] Gruber, A., Su, C., Zwieback, S., Crow, W., Dorigo, W.,\
    Wagner, W.  (2015). Recent advances in (soil moisture) triple\
    collocation analysis.  International Journal of Applied Earth\
    Observation and Geoinformation, in review
    """

    cov = np.cov(np.vstack((x, y, z)))
    return _tcol_metrics_from_cov(cov, ref_ind)


def _tcol_metrics_from_cov(cov, ref_ind=0):
    """
    Calculate TCA metrics from pre-computed covariance matrix
    """
    ind = (0, 1, 2, 0, 1, 2)
    no_ref_ind = np.where(np.arange(3) != ref_ind)[0]

    # we have to take the absolute value of the ratio of covariances and of the
    # full difference according to Gruber et al. 2020, Eq. 11
    snr = 10 * np.log10(
        [
            np.abs(
                np.abs((cov[i, i] * cov[ind[i + 1], ind[i + 2]])
                       / (cov[i, ind[i + 1]] * cov[i, ind[i + 2]]))
                - 1
            )
            ** (-1)
            for i in np.arange(3)
        ]
    )
    err_var = np.array(
        [
            cov[i, i]
            - (cov[i, ind[i + 1]] * cov[i, ind[i + 2]])
            / cov[ind[i + 1], ind[i + 2]]
            for i in np.arange(3)
        ]
    )

    beta = np.array(
        [
            cov[ref_ind, no_ref_ind[no_ref_ind != i][0]]
            / cov[i, no_ref_ind[no_ref_ind != i][0]]
            if i != ref_ind
            else 1
            for i in np.arange(3)
        ]
    )

    return snr, np.sqrt(err_var) * beta, beta


def check_if_biased(combs, correlated):
    """
    Supporting function for extended collocation
    Checks whether the estimators are biased by checking of not
    too manny data sets (are assumed to have cross-correlated errors)
    """
    for corr in correlated:
        for comb in combs:
            if (np.array_equal(comb, corr)) | (
                np.array_equal(comb, corr[::-1])
            ):
                return True
    return False


def ecol(data, correlated=None, err_cov=None, abs_est=True):
    """
    Extended collocation analysis to obtain estimates of:

        - signal variances
        - error variances
        - signal-to-noise ratios [dB]
        - error cross-covariances (and -correlations)

    based on an arbitrary number of N>3 data sets.

    **EACH DATA SET MUST BE MEMBER OF >= 1 TRIPLET THAT FULFILLS THE CLASSICAL
    TRIPLE COLLOCATION ASSUMPTIONS**

    Parameters
    ----------
    data : pd.DataFrame
        Temporally matched input data sets in each column
    correlated : tuple of tuples (string)
        A tuple containing tuples of data set names (column names), between
        which the error cross-correlation shall be estimated.
        e.g. [['AMSR-E','SMOS'],['GLDAS','ERA']] estimates error
        cross-correlations between (AMSR-E and SMOS), and (GLDAS and ERA),
        respectively.
    err_cov :
        A priori known error cross-covariances that shall be included
        in the estimation (to obtain unbiased estimates)
    abs_est :
        Force absolute values for signal and error variance estimates
        (to mitiate the issue of estimation uncertainties)

    Returns
    -------
    A dictionary with the following entries (<name> correspond to data set (df
    column's) names:
    - sig_<name> : signal variance of <name>
    - err_<name> : error variance of <name>
    - snr_<name> : SNR (in dB) of <name>
    - err_cov_<name1>_<name2> : error covariance between <name1> and <name2>
    - err_corr_<name1>_<name2> : error correlation between <name1> and <name2>

    Notes
    -----
    Rescaling parameters can be derived from the signal variances
    e.g., scaling <src> against <ref>:
    beta =  np.sqrt(sig_<ref> / sig_<src>)
    rescaled = (data[<src>] - data[<src>].mean()) * beta + data[<ref>].mean()

    Examples
    --------
    .. code-block::

        # Just random numbers for demonstrations
        ds1 = np.random.normal(0,1,500)
        ds2 = np.random.normal(0,1,500)
        ds3 = np.random.normal(0,1,500)
        ds4 = np.random.normal(0,1,500)
        ds5 = np.random.normal(0,1,500)

        # Three data sets without cross-correlated errors: This is equivalent
        # to standard triple collocation.
        df = pd.DataFrame({'ds1':ds1,'ds2':ds2,'ds3':ds3},
                          index=np.arange(500))
        res = ecol(df)

        # Five data sets, where data sets (1 and 2), and (3 and 4), are assumed
        # to have cross-correlated errors.
        df = pd.DataFrame({'ds1':ds1,'ds2':ds2,'ds3':ds3,'ds4':ds4,'ds5':ds5},
                          index=np.arange(500),)
        correlated = [['ds1','ds2'],['ds3','ds4']]
        res = ecol(df,correlated=correlated)

    References
    ----------
    .. [Gruber2016] Gruber, A., Su, C. H., Crow, W. T., Zwieback, S., Dorigo,\
    W. A., & Wagner, W. (2016). Estimating error cross-correlations in soil\
    moisture data sets using extended collocation analysis. Journal of\
    Geophysical Research: Atmospheres, 121(3), 1208-1219.
    """
    data.dropna(inplace=True)

    cols = data.columns.values
    cov = data.cov()

    # subtract a-priori known error covariances to obtained unbiased estimates
    if err_cov is not None:
        cov[err_cov[0]][err_cov[1]] -= err_cov[2]
        cov[err_cov[1]][err_cov[0]] -= err_cov[2]

    # ----- Building up the observation vector y and the design matrix A -----

    # Initial lenght of the parameter vector x:
    # n error variances + n signal variances
    n_x = 2 * len(cols)

    # First n elements in y: variances of all data sets
    y = cov.values.diagonal()

    # Extend y if data sets with correlated errors exist
    if correlated is not None:
        # additionally estimated in x:
        # k error covariances, and k cross-biased signal variances
        # (biased with the respective beta_i*beta_j)
        n_x += 2 * len(correlated)

        # add covariances between the correlated data sets to the y vector
        y = np.hstack((y, [cov[ds[0]][ds[1]] for ds in correlated]))

    # Generate the first part of the design matrix A (total variance = signal
    # variance + error variance)
    A = np.hstack(
        (
            np.matrix(np.identity(int(n_x / 2))),
            np.matrix(np.identity(int(n_x / 2))),
        )
    ).astype("int")

    # build up A and y components for estimating signal variances (biased with
    # beta_i**2 only)
    # i.e., the classical TC based signal variance estimators
    # cov(a,c)*cov(a,d)/cov(c,d)
    for col in cols:

        others = cols[cols != col]
        combs = list(combinations(others, 2))

        for comb in combs:
            if correlated is not None:
                if check_if_biased(
                    [[col, comb[0]], [col, comb[1]], [comb[0], comb[1]]],
                    correlated,
                ):
                    continue

            A_line = np.zeros(n_x).astype("int")
            A_line[np.where(cols == col)[0][0]] = 1
            A = np.vstack((A, A_line))

            y = np.append(
                y,
                cov[col][comb[0]] * cov[col][comb[1]] / cov[comb[0]][comb[1]],
            )

    # build up A and y components for the cross-biased signal variabilities
    # (with beta_i*beta_j)
    # i.e., the cross-biased signal variance estimators
    # (cov(a,c)*cov(b,d)/cov(c,d))
    if correlated is not None:
        for i in np.arange(len(correlated)):
            others = cols[
                (cols != correlated[i][0]) & (cols != correlated[i][1])
            ]
            combs = list(permutations(others, 2))

            for comb in combs:
                if check_if_biased(
                    [
                        [correlated[i][0], comb[0]],
                        [correlated[i][1], comb[1]],
                        comb,
                    ],
                    correlated,
                ):
                    continue

                A_line = np.zeros(n_x).astype("int")
                A_line[len(cols) + i] = 1
                A = np.vstack((A, A_line))

                y = np.append(
                    y,
                    cov[correlated[i][0]][comb[0]]
                    * cov[correlated[i][1]][comb[1]]
                    / cov[comb[0]][comb[1]],
                )

    y = np.matrix(y).T

    # ----- Solving for the parameter vector x -----

    x = (A.T * A).I * A.T * y
    x = np.squeeze(np.array(x))

    # ----- Building up the result dictionary -----

    tags = np.hstack(("sig_" + cols, "err_" + cols, "snr_" + cols))

    if correlated is not None:

        # remove the cross-biased signal variabilities (with beta_i*beta_j) as
        # they are not useful
        x = np.delete(x, np.arange(len(correlated)) + len(cols))

        # Derive error cross-correlations from error covariances and error
        # variances
        for i in np.arange(len(correlated)):
            x = np.append(
                x,
                x[2 * len(cols) + i]
                / np.sqrt(
                    x[len(cols) + np.where(cols == correlated[i][0])[0][0]]
                    * x[len(cols) + np.where(cols == correlated[i][1])[0][0]]
                ),
            )

        # add error covariances and correlations to the result dictionary
        tags = np.append(
            tags,
            np.array(["err_cov_" + ds[0] + "_" + ds[1] for ds in correlated]),
        )
        tags = np.append(
            tags,
            np.array(["err_corr_" + ds[0] + "_" + ds[1] for ds in correlated]),
        )

    # force absolute signal and error variance estimates to compensate for
    # estimation uncertainties
    if abs_est is True:
        x[0:(2 * len(cols))] = np.abs(x[0:(2 * len(cols))])

    # calculate and add SNRs (in decibel units)
    x = np.insert(
        x,
        2 * len(cols),
        10 * np.log10(x[0:len(cols)] / x[len(cols):(2 * len(cols))]),
    )

    return dict(zip(tags, x))
